from __future__ import annotations

from typing import Dict, Union, List, Any

from lxml import etree as et

from xml_common.utils import Language
from eml.types import EMLObject, Scope, I18nString


class EMLAddress(EMLObject):
    """
    The full address information for a given responsible party entry.

    Parameters
    ----------
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined
    references_system : str, optional
        System attribute of reference
    delivery_point : str, I18nString or List[I18String], optional
        The location for postal deliveries.
    city : str or I18nString, optional
        The name of the city for the contact.
    administrative_area : str or I18nString, optional
        The political area of a country.
    postal_code : str or I18nString, optional
        The postal code used for routing to an address.
    country : str or I18nString, optional
        The name of the country for the contact's address.
    """
    PRINCIPAL_TAG = "address"

    def __init__(
            self, _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
            referencing: bool = False,
            references_system: str = None,
            delivery_point: Union[str, I18nString, List[I18nString]] = None,
            city: Union[str, I18nString] = None,
            administrative_area: Union[str, I18nString] = None,
            postal_code: Union[str, I18nString] = None,
            country: Union[str, I18nString] = None,
            language: Language.ENG = None,
    ) -> None:
        super().__init__(_id, scope, system, referencing, references_system)
        self.__delivery_point__: List[I18nString] = list()
        if isinstance(delivery_point, list):
            self.__delivery_point__.extend(delivery_point)
        elif delivery_point is not None:
            self.__delivery_point__.append(I18nString(delivery_point, lang=language))
        self.__city__ = None
        if city is not None:
            self.__city__ = I18nString(city, lang=language)
        self.__admin_area__ = None
        if administrative_area is not None:
            self.__admin_area__ = I18nString(administrative_area, lang=language)
        self.__postal_code__ = None
        if postal_code is not None:
            self.__postal_code__ = I18nString(postal_code, lang=language)
        self.__country__ = None
        if country is not None:
            self.__country__ = I18nString(country, lang=language)
        return

    @property
    def delivery_point(self) -> List[I18nString]:
        """
        List[I18nString]: The location for postal deliveries.
        """
        return self.__delivery_point__

    @property
    def city(self) -> I18nString:
        """
        I18String: The name of the city for the contact.
        """
        return self.__city__

    @property
    def administrative_area(self) -> I18nString:
        """
        I18nString: The political area of a country.
        """
        return self.__admin_area__

    @property
    def postal_code(self) -> I18nString:
        """
        I18nString: The postal code used for routing to an address.
        """
        return self.__postal_code__

    @property
    def country(self) -> I18nString:
        """
        I18nString: The name of the country for the contact's address.
        """
        return self.__country__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLAddress:
        """
        Generate an EML Address referencing another EML Address.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references EML address.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLAddress
            Object parsed that reference an Address.
        """
        references = element.find("references")
        address = EMLAddress(
            _id=references.text,
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None)
        )
        return address

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLAddress:
        """
        Generate an EML Object that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLAddress
            Object parsed.
        """
        delivery_points = list()
        for delivery_point in element.findall("deliveryPoint"):
            delivery_points.append(I18nString.parse(delivery_point, nmap))
        address = EMLAddress(
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            delivery_point=delivery_points,
            city=I18nString.parse(element.find("city", nmap), nmap),
            administrative_area=I18nString.parse(element.find("administrativeArea", nmap), nmap),
            postal_code=I18nString.parse(element.find("postalCode", nmap), nmap),
            country=I18nString.parse(element.find("country", nmap), nmap)
        )
        return address

    def to_element(self) -> et.Element:
        """
        Generate a lxml.tree.Element from the object.

        Returns
        -------
        lxml.tree.Element
            Object in the Element format.
        """
        address = super().to_element()
        address = self._to_element_(address)
        references = self.generate_references_element()
        if references is None:
            for delivery_point in self.delivery_point:
                delivery_point.set_tag("deliveryPoint")
                address.append(delivery_point.to_element())
            if self.city is not None:
                self.city.set_tag("city")
                address.append(self.city.to_element())
            if self.administrative_area:
                self.administrative_area.set_tag("administrativeArea")
                address.append(self.administrative_area.to_element())
            if self.postal_code is not None:
                self.postal_code.set_tag("postalCode")
                address.append(self.postal_code.to_element())
            if self.country is not None:
                self.country.set_tag("country")
                address.append(self.country.to_element())
        return address

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, EMLAddress):
            return (super().__eq__(other) and
                    sorted(self.delivery_point) == sorted(other.delivery_point) and
                    self.city == other.city and
                    self.administrative_area == other.administrative_area and
                    self.postal_code == other.postal_code and
                    self.country == other.country)
        else:
            return False



import unittest

import datetime as dt
from lxml import etree as et

from eml.resources.coverage import TemporalCoverage
from eml.types import Scope
from test_xml.test_xml import TestXML


class TestTemporalCoverage(TestXML):
    def test_parse_none(self):
        self.assertIsNone(TemporalCoverage.parse(None, {}), "Time Coverage from nowhere")

    def test_parse_invalid(self):
        text_xml = """
<temporalCoverage id="1" scope="system" system="http://gbif.org">
</temporalCoverage>
        """
        self.assertRaises(
            ValueError,
            TemporalCoverage.from_string,
            text_xml
        )

    def test_parse(self):
        text_xml = """
<temporalCoverage id="1" scope="system" system="http://gbif.org">
    <singleDateTime>
        <calendarDate>2001-01-01</calendarDate>
        <time>14:06:09-08:00</time>
    </singleDateTime>
</temporalCoverage>
        """
        coverage = TemporalCoverage.from_string(text_xml)
        self.assertEqual("1", coverage.id, "Error on parsing id")
        self.assertEqual(Scope.SYSTEM, coverage.scope, "Error on parsing scope")
        self.assertFalse(coverage.referencing, "Error on parsing")
        self.assertEqual("http://gbif.org", coverage.system, "Error on parsing system")
        self.assertEqual(
            dt.datetime(
                2001, 1, 1,
                14, 6, 9,
                tzinfo=dt.timezone(dt.timedelta(hours=-8, minutes=0))
            ),
            coverage.single_datetime[0],
            "Error on parsing single date time"
        )
        self.assertEqual(1, len(coverage.single_datetime), "Error on parsing single date time")
        self.assertIsNone(coverage.range_datetime, "Range datetime from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), coverage.to_element(), "Error on to element")

    def test_parse_multiple_dates(self):
        text_xml = """
<temporalCoverage id="1" scope="system" system="http://gbif.org">
    <singleDateTime>
        <calendarDate>2001-01-01</calendarDate>
        <time>14:06:09-08:00</time>
    </singleDateTime>
    <singleDateTime>
        <calendarDate>2011-12-31</calendarDate>
        <time>08:31:22Z</time>
    </singleDateTime>
</temporalCoverage>
        """
        coverage = TemporalCoverage.from_string(text_xml)
        self.assertEqual("1", coverage.id, "Error on parsing id")
        self.assertEqual(Scope.SYSTEM, coverage.scope, "Error on parsing scope")
        self.assertFalse(coverage.referencing, "Error on parsing")
        self.assertEqual("http://gbif.org", coverage.system, "Error on parsing system")
        self.assertEqual(2, len(coverage.single_datetime), "Error on parsing single date time")
        self.assertEqual(
            dt.datetime(
                2001, 1, 1,
                14, 6, 9,
                tzinfo=dt.timezone(dt.timedelta(hours=-8))
            ),
            coverage.single_datetime[0],
            "Error on parsing single date time"
        )
        self.assertEqual(
            dt.datetime(
                2011, 12, 31,
                8, 31, 22,
                tzinfo=dt.timezone.utc
            ),
            coverage.single_datetime[1],
            "Error on parsing second single date time"
        )
        self.assertIsNone(coverage.range_datetime, "Range datetime from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), coverage.to_element(), "Error on to element")

    def test_parse_range(self):
        text_xml = """
<temporalCoverage id="1" scope="system" system="http://gbif.org">
    <rangeOfDates>
        <beginDate>
            <calendarDate>2001-01-01</calendarDate>
            <time>08:31:22Z</time>
        </beginDate>
        <endDate>
            <calendarDate>2021</calendarDate>
        </endDate>
    </rangeOfDates>
</temporalCoverage>
        """
        coverage = TemporalCoverage.from_string(text_xml)
        self.assertEqual("1", coverage.id, "Error on parsing id")
        self.assertEqual(Scope.SYSTEM, coverage.scope, "Error on parsing scope")
        self.assertFalse(coverage.referencing, "Error on parsing")
        self.assertEqual("http://gbif.org", coverage.system, "Error on parsing system")
        self.assertEqual(0, len(coverage.single_datetime), "Single datetime from nowhere")
        self.assertEqual(
            dt.datetime(
                2001, 1, 1,
                8, 31, 22,
                tzinfo=dt.timezone.utc
            ),
            coverage.range_datetime[0],
            "Error on parsing begin range date time"
        )
        self.assertEqual(
            dt.date(
                2021, 1, 1
            ),
            coverage.range_datetime[1],
            "Error on parsing end range date time"
        )
        self.assertEqualTree(et.fromstring(text_xml), coverage.to_element(), "Error on to element")

    def test_parse_alt_time_scale_none(self):
        self.assertIsNone(
            TemporalCoverage.AlternativeTimeScale.parse(
            None, {}),
            "Alterative from nowhere"
        )

    def test_parse_alternative_time_scale(self):
        wikipedia_explanation = ("Through analysis of seafloor magnetic "
                                 "anomalies and dating of reversal "
                                 "sequences on land, paleomagnetists "
                                 "have been developing a Geomagnetic "
                                 "Polarity Time Scale (GPTS).")
        text_xml = f"""
<alternativeTimeScale>
    <timeScaleName>Geomagnetic Polarity Time Scale</timeScaleName>
    <timeScaleAgeEstimate>C28r</timeScaleAgeEstimate>
    <timeScaleAgeUncertainty>+/- 5 Ma</timeScaleAgeUncertainty>
    <timeScaleAgeExplanation>{wikipedia_explanation}</timeScaleAgeExplanation>
    <timeScaleCitation scope="document">
        <references>3</references>
    </timeScaleCitation>
</alternativeTimeScale>
        """
        alternative_ts = TemporalCoverage.AlternativeTimeScale.from_string(text_xml)
        self.assertEqual(
            "Geomagnetic Polarity Time Scale",
            alternative_ts.name,
            "Error on parsing time scale name"
        )
        self.assertEqual(
            "C28r",
            alternative_ts.age_estimate,
            "Error on parsing time scale age estimate"
        )
        self.assertEqual(
            "+/- 5 Ma",
            alternative_ts.age_uncertainty,
            "Error on parsing time scale age uncertainty"
        )
        self.assertEqual(
            wikipedia_explanation,
            alternative_ts.age_explanation,
            "Error on parsing time scale age explanation"
        )
        self.assertEqual(1, len(alternative_ts.citation), "Error on parsing time scale citation")
        self.assertEqual("3", alternative_ts.citation[0].id, "Error on parsing time scale citation")
        self.assertTrue(alternative_ts.citation[0].referencing, "Error on parsing time scale citation")
        self.assertEqualTree(et.fromstring(text_xml), alternative_ts.to_element(), "Error on to element")

    def test_parse_alternative_time(self):
        text_xml = f"""
<temporalCoverage id="1" scope="system" system="http://gbif.org">
    <singleDateTime>
        <alternativeTimeScale>
            <timeScaleName>Geomagnetic Polarity Time Scale</timeScaleName>
            <timeScaleAgeEstimate>C28r</timeScaleAgeEstimate>
            <timeScaleCitation scope="document">
                <references>3</references>
            </timeScaleCitation>
        </alternativeTimeScale>
    </singleDateTime>
</temporalCoverage>
        """
        coverage = TemporalCoverage.from_string(text_xml)
        self.assertEqual("1", coverage.id, "Error on parsing id")
        self.assertEqual(Scope.SYSTEM, coverage.scope, "Error on parsing scope")
        self.assertFalse(coverage.referencing, "Error on parsing")
        self.assertEqual("http://gbif.org", coverage.system, "Error on parsing system")
        self.assertEqual(1, len(coverage.single_datetime), "Error on parsing single datetime")
        self.assertEqual(
            "Geomagnetic Polarity Time Scale",
            coverage.single_datetime[0].name,
            "Error on parsing time scale name"
        )
        self.assertEqual(
            "C28r",
            coverage.single_datetime[0].age_estimate,
            "Error on parsing time scale age estimate"
        )
        self.assertIsNone(
            coverage.single_datetime[0].age_uncertainty,
            "Time scale age uncertainty from nowhere"
        )
        self.assertIsNone(
            coverage.single_datetime[0].age_explanation,
            "Time scale age explanation from nowhere"
        )
        self.assertEqual(1, len(coverage.single_datetime[0].citation), "Error on parsing time scale citation")
        self.assertEqual("3", coverage.single_datetime[0].citation[0].id, "Error on parsing time scale citation")
        self.assertTrue(coverage.single_datetime[0].citation[0].referencing, "Error on parsing time scale citation")
        self.assertEqualTree(et.fromstring(text_xml), coverage.to_element(), "Error on to element")

    def test_parse_reference(self):
        text_xml = """
<temporalCoverage scope="system">
    <references system="http://gbif.org">1</references>
</temporalCoverage>
        """
        coverage = TemporalCoverage.from_string(text_xml)
        self.assertEqual("1", coverage.id, "Error on parsing id")
        self.assertEqual(Scope.SYSTEM, coverage.scope, "Error on parsing scope")
        self.assertIsNone(coverage.system, "System from nowhere")
        self.assertTrue(coverage.referencing, "Error on parsing")
        self.assertEqual("http://gbif.org", coverage.references.system, "Error on parsing references")
        self.assertIsNone(coverage.range_datetime, "Range datetime from nowhere")
        self.assertEqual(0, len(coverage.single_datetime), "Single datetime from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), coverage.to_element(), "Error on to element")

    def test_not_valid_datetime_to_xml(self):
        self.assertRaises(
            TypeError,
            TemporalCoverage.datetime_to_xml,
            "2001-01-01"
        )
        self.assertRaises(
            TypeError,
            TemporalCoverage.datetime_to_xml,
            (2001, 1, 1)
        )
        self.assertRaises(
            TypeError,
            TemporalCoverage.datetime_to_xml,
            2021
        )


if __name__ == '__main__':
    unittest.main()



import unittest

from lxml import etree as et

from dwca.terms import OrganismQuantity
from test_xml.test_xml import TestXML


class TestUnionField(TestXML):
    def test_parse(self):
        organism_quantity_xml = """<field index="3" term="http://rs.tdwg.org/dwc/terms/organismQuantity"/>"""
        organism_quantity = OrganismQuantity.from_string(organism_quantity_xml)
        self.assertEqual(
            "http://rs.tdwg.org/dwc/terms/organismQuantity",
            organism_quantity.uri,
            "Wrong URI"
        )
        self.assertEqual(3, organism_quantity.index, "Wrong index")
        self.assertEqualTree(
            et.fromstring(organism_quantity_xml),
            organism_quantity.to_element(),
            "Wrong generated element."
        )
        pass

    def test_format(self):
        organism_quantity = OrganismQuantity(3)
        value = "many"
        self.assertEqual(
            value, organism_quantity.format(value), "Error formatting union (string)"
        )
        value = "14.8"
        self.assertGreaterEqual(
            1e-10, abs(organism_quantity.format(value) - 14.8), "Error formatting union (float)"
        )
        value = "3"
        self.assertEqual(
            3, organism_quantity.format(value), "Error formatting union (int)"
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from enum import Enum
from typing import Dict, List, Tuple, Any

from lxml import etree as et

from xml_common import XMLObject
from eml.types import EMLObject, Scope, ExtensionString, EMLTextType


class EMLOnline(XMLObject):
    """
    Online distribution information.

    Parameters
    ----------
    online_description : I18nString or str, optional
        Brief description of the content of online.
    url : str, optional
        A URL from which this resource can be downloaded or information can be obtained about downloading it.
    url_function : EMLOnline.FunctionType, optional
        "information" when the URL meant for informational purposes or "download" URL returns the data stream itself.
    connection : EMLOnline.Connection, optional
         A connection to a data service.
    connection_definition : EMLOnline.ConnectionDefinition, optional
        The definition of a connection that will be used in another location in the EML document.
    """
    PRINCIPAL_TAG = "online"

    class FunctionType(Enum):
        """
        Function of the URL.
        """
        DOWNLOAD = 0
        INFORMATION = 1

    class ParameterDefinition(XMLObject):
        """
        The definition of a parameter that is needed to properly use this connection scheme.

        Parameters
        ----------
        name : str
            The Name of a parameter that is needed to properly use this connection scheme.
        definition : str
            The definition of a parameter that is needed to properly use this connection scheme.
        default_value : str, optional
            The default value for a parameter that is needed to properly use this connection scheme.
        """
        PRINCIPAL_TAG = "parameterDefinition"
        """str: Principal tag `parameterDefinition`"""

        def __init__(self, name: str, definition: str, default_value: str = None) -> None:
            super().__init__()
            self.__name__ = name
            self.__def__ = definition
            self.__default__ = default_value
            return

        @property
        def name(self) -> str:
            """str: The Name of a parameter that is needed to properly use this connection scheme."""
            return self.__name__

        @property
        def definition(self) -> str:
            """str: The definition of a parameter that is needed to properly use this connection scheme."""
            return self.__def__

        @property
        def default_value(self) -> str:
            """str: The default value for a parameter that is needed to properly use this connection scheme."""
            return self.__default__

        @classmethod
        def parse(cls, element: et.Element, nmap: Dict) -> EMLOnline.ParameterDefinition | None:
            """
            Parse an XML element into a ParameterDefinition object

            Parameters
            ----------
            element : xml.etree.Element
                XML element to be parsed
            nmap : Dict
                Namespace

            Returns
            -------
            EMLOnline.ParameterDefinition
                Instance with the information of the XML element
            """
            name_elem = element.find("name", nmap)
            def_elem = element.find("definition", nmap)
            default_elem = element.find("defaultValue", nmap)
            default_value = None
            if default_elem is not None:
                default_value = default_elem.text if default_elem.text is not None else ""
            param_def = EMLOnline.ParameterDefinition(
                name=name_elem.text if name_elem.text is not None else "",
                definition=def_elem.text if def_elem.text is not None else "",
                default_value=default_value
            )
            param_def.__namespace__ = nmap
            return param_def

        def to_element(self) -> et.Element:
            """
            Generate an XML element from Parameter Definition instance

            Returns
            -------
            xml.etree.Element
                XML element instance with the Parameter Definition information
            """
            param_def = super().to_element()
            name_elem = self.object_to_element("name")
            name_elem.text = self.name
            param_def.append(name_elem)
            def_elem = self.object_to_element("definition")
            def_elem.text = self.definition
            param_def.append(def_elem)
            if self.default_value is not None:
                default_elem = self.object_to_element("defaultValue")
                default_elem.text = self.default_value
                param_def.append(default_elem)
            return param_def

        def __eq__(self, other: Any) -> bool:
            if isinstance(other, EMLOnline.ParameterDefinition):
                return (
                    self.name == other.name and
                    self.definition == other.definition
                )
            else:
                return False

        def __gt__(self, other: EMLOnline.ParameterDefinition) -> bool:
            if not isinstance(other, EMLOnline.ParameterDefinition):
                raise TypeError(f"'>' not supported between instances of 'EML.ParameterDefinition' and '{type(other)}'")
            else:
                return self.name > other.name

        def __ge__(self, other: EMLOnline.ParameterDefinition) -> bool:
            return self > other or self == other

        def __lt__(self, other: EMLOnline.ParameterDefinition) -> bool:
            if not isinstance(other, EMLOnline.ParameterDefinition):
                raise TypeError(f"'<' not supported between instances of 'EML.ParameterDefinition' and '{type(other)}'")
            else:
                return self.name < other.name

        def __le__(self, other: EMLOnline.ParameterDefinition) -> bool:
            return self < other or self == other

    class ConnectionDefinition(EMLObject):
        """
        Definition of the connection protocol.

        Parameters
        ----------
        _id : str, optional
            Unique identifier within the scope.
        scope : Scope, default DOCUMENT
            The scope of the identifier.
        system : str, optional
            The data management system within which an identifier is in scope and therefore unique.
        referencing : bool, optional, default=False
            Whether the resource is referencing another or is being defined.
        references_system : str, optional
            System attribute of reference.
        scheme_name : ExtensionString, optional
            The name of the scheme used to identify this connection.
        description : EMLTextType, optional
            The description of the scheme used to identify this connection.
        parameters : List[EMLOnline.ParameterDefinition], optional
            A list of definitions of a parameter that is needed to properly use this connection scheme.
        """
        PRINCIPAL_TAG = "connectionDefinition"
        """str: Principal tag `connectionDefinition`"""

        def __init__(
                self, _id: str = None,
                scope: Scope = Scope.DOCUMENT,
                system: str = None,
                referencing: bool = False,
                references_system: str = None,
                scheme_name: ExtensionString = None,
                description: EMLTextType = None,
                parameters: List[EMLOnline.ParameterDefinition] = None,
        ) -> None:
            super().__init__(_id, scope, system, referencing, references_system)
            self.__scheme_name__ = scheme_name
            self.__description__ = description
            self.__param__: List[EMLOnline.ParameterDefinition] = list()
            if parameters is not None:
                self.__param__.extend(parameters)
            return

        @property
        def scheme_name(self) -> ExtensionString:
            """ExtensionString: The name of the scheme used to identify this connection."""
            return self.__scheme_name__

        @property
        def description(self) -> EMLTextType:
            """EMLTextType: The description of the scheme used to identify this connection."""
            return self.__description__

        @property
        def parameters(self) -> List[EMLOnline.ParameterDefinition]:
            """
            List[EMLOnline.ParameterDefinition]: List of parameters that are needed to use this connection scheme.
            """
            return self.__param__

        @classmethod
        def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLOnline.ConnectionDefinition:
            """
            Generate a ConnectionDefinition object referencing another from an XML element

            Parameters
            ----------
            element : lxml.etree.Element
                XML element object to parse
            nmap : Dict
                Namespace

            Returns
            -------
            EMLOnline.ConnectionDefinition
                Instance parsed from XML Element
            """
            references = element.find("references", nmap)
            return EMLOnline.ConnectionDefinition(
                _id=references.text,
                scope=cls.get_scope(element),
                system=element.get("system", None),
                referencing=True,
                references_system=references.get("system", None),
            )

        @classmethod
        def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLOnline.ConnectionDefinition:
            """
            Generate a ConnectionDefinition object from an XML element

            Parameters
            ----------
            element : lxml.etree.Element
                XML element object to parse
            nmap : Dict
                Namespace

            Returns
            -------
            EMLOnline.ConnectionDefinition
                Instance parsed from XML Element
            """
            scheme_elem = element.find("schemeName", nmap)
            if scheme_elem is not None:
                scheme_name = ExtensionString.parse(scheme_elem, nmap)
            else:
                scheme_name = None
            des_elem = element.find("description", nmap)
            if des_elem is not None:
                description = EMLTextType.parse(des_elem, nmap)
            else:
                description = None
            parameters = list()
            for param_elem in element.findall("parameterDefinition", nmap):
                parameters.append(EMLOnline.ParameterDefinition.parse(param_elem, nmap))
            return EMLOnline.ConnectionDefinition(
                _id=element.get("id", None),
                scope=cls.get_scope(element),
                system=element.get("system", None),
                referencing=False,
                scheme_name=scheme_name,
                description=description,
                parameters=parameters,
            )

        def to_element(self) -> et.Element:
            """
            Generate an XML element with the instance information.

            Returns
            -------
            lxml.etree.Element
                XML Element instance
            """
            conn_def = super().to_element()
            conn_def = self._to_element_(conn_def)
            if self.referencing:
                return conn_def
            if self.scheme_name is not None:
                self.scheme_name.set_tag("schemeName")
                conn_def.append(self.scheme_name.to_element())
            if self.description is not None:
                self.description.set_tag("description")
                conn_def.append(self.description.to_element())
            for parameter in self.parameters:
                conn_def.append(parameter.to_element())
            return conn_def

        def __eq__(self, other: Any) -> bool:
            if isinstance(other, EMLOnline.ConnectionDefinition):
                parameter_equals = self.parameters is None and other.parameters is None
                if not parameter_equals:
                    if self.parameters is not None and other.parameters is not None:
                        parameter_equals = sorted(self.parameters) == sorted(other.parameters)
                return (
                    super().__eq__(other) and
                    self.scheme_name == other.scheme_name and
                    self.description == other.description and
                    parameter_equals
                )
            else:
                return False

    class Connection(EMLObject):
        """
        A description of the information needed to make an application connection to a data service.

        Parameters
        ----------
        _id : str, optional
            Unique identifier within the scope.
        scope : Scope, default DOCUMENT
            The scope of the identifier.
        system : str, optional
            The data management system within which an identifier is in scope and therefore unique.
        referencing : bool, optional, default=False
            Whether the resource is referencing another or is being defined.
        references_system : str, optional
            System attribute of reference.
        connection_definition : EMLOnline.ConnectionDefinition, optional
            Definition of the connection protocol.
        parameters : List[Tuple[str, str]], optional
            A parameter to be used to make this connection.
        """
        PRINCIPAL_TAG = "connection"

        def __init__(
                self, _id: str = None,
                scope: Scope = Scope.DOCUMENT,
                system: str = None,
                referencing: bool = False,
                references_system: str = None,
                connection_definition: EMLOnline.ConnectionDefinition = None,
                parameters: List[Tuple[str, str]] = None,
        ) -> None:
            super().__init__(_id, scope, system, referencing, references_system)
            self.__conn_def__ = connection_definition
            self.__param__ = list()
            if parameters is not None:
                self.__param__.extend(parameters)
            return

        @property
        def connection_definition(self) -> EMLOnline.ConnectionDefinition:
            """
            EMLOnline.ConnectionDefinition: Definition of the connection protocol.
            """
            return self.__conn_def__

        @property
        def parameters(self) -> List[Tuple[str, str]]:
            """List[Tuple[str, str]]: A parameter to be used to make this connection."""
            return self.__param__

        @classmethod
        def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLOnline.Connection:
            """
            Generate a Connection instance referencing another instance.

            Parameters
            ----------
            element : lxml.etree.Element
                An XML element instance to parse.
            nmap : Dict
                Namespace.

            Returns
            -------
            EMLOnline.Connection
                Instance of Connection referencing another.
            """
            references = element.find("references", nmap)
            return EMLOnline.Connection(
                _id=references.text,
                scope=cls.get_scope(element),
                system=element.get("system", None),
                referencing=True,
                references_system=references.get("system", None)
            )

        @classmethod
        def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLOnline.Connection:
            """
            Generate a Connection instance from an XML element.

            Parameters
            ----------
            element : lxml.etree.Element
                An XML element instance to parse.
            nmap : Dict
                Namespace.

            Returns
            -------
            EMLOnline.Connection
                Instance with the information from the XML element.
            """
            des_elem = element.find("connectionDefinition", nmap)
            if des_elem is not None:
                conn_des = EMLOnline.ConnectionDefinition.parse(des_elem, nmap)
            else:
                conn_des = None
            parameters = list()
            for param_elem in element.findall("parameter", nmap):
                parameters.append((
                    param_elem.find("name").text,
                    param_elem.find("value").text,
                ))
            return EMLOnline.Connection(
                _id=element.get("id", None),
                scope=cls.get_scope(element),
                system=element.get("system", None),
                referencing=False,
                connection_definition=conn_des,
                parameters=parameters,
            )

        def to_element(self) -> et.Element:
            """
            Generate an XML Element with instance information.

            Returns
            -------
            lxml.etree.Element
                An XML Element instance.
            """
            conn_elem = super().to_element()
            conn_elem = self._to_element_(conn_elem)
            if self.referencing:
                return conn_elem
            if self.connection_definition is not None:
                conn_elem.append(self.connection_definition.to_element())
            for parameter in self.parameters:
                param_elem = self.object_to_element("parameter")
                name_elem = self.object_to_element("name")
                name_elem.text = parameter[0]
                param_elem.append(name_elem)
                value_elem = self.object_to_element("value")
                value_elem.text = parameter[1]
                param_elem.append(value_elem)
                conn_elem.append(param_elem)
            return conn_elem

        def __eq__(self, other: Any) -> bool:
            if isinstance(other, EMLOnline.Connection):
                return (
                    super().__eq__(other) and
                    self.connection_definition == other.connection_definition and
                    sorted(self.parameters) == sorted(other.parameters)
                )
            else:
                return False

    def __init__(
            self, online_description: str = None,
            url: str = None,
            url_function: EMLOnline.FunctionType = None,
            connection: EMLOnline.Connection = None,
            connection_definition: EMLOnline.ConnectionDefinition = None
    ) -> None:
        super().__init__()
        conn = 0
        if url is not None:
            conn += 1
        if connection is not None:
            conn += 1
        if connection_definition is not None:
            conn += 1
        if conn > 1:
            raise ValueError(
                "Just one of the following can be given: "
                "`online_description`, `connection` or "
                "`connection_definition`"
            )
        self.__online_description__ = online_description
        self.__url__ = url
        if self.__url__ is not None:
            self.__url_func__ = EMLOnline.FunctionType.DOWNLOAD
        else:
            self.__url_func__ = None
        if url_function is not None:
            self.__url_func__ = url_function
        self.__conn__ = connection
        self.__conn_def__ = connection_definition
        return

    @property
    def online_description(self) -> str:
        """str: Brief description of the content of online."""
        return self.__online_description__

    @property
    def url(self) -> str:
        """str: A URL from which this resource can be downloaded or information can be obtained about downloading it."""
        return self.__url__

    @property
    def url_function(self) -> EMLOnline.FunctionType:
        """EMLOnline.FunctionType: Function of the url."""
        return self.__url_func__

    @property
    def connection(self) -> EMLOnline.Connection:
        """EMLOnline.Connection: A connection to a data service."""
        return self.__conn__

    @property
    def connection_definition(self) -> EMLOnline.ConnectionDefinition:
        """
        EMLOnline.ConnectionDefinition: Definition of the connection protocol.
        """
        return self.__conn_def__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLOnline | None:
        """
        Generate an EMLOnline instance from an XML element.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLOnline
            Instance parsed.
        """
        if element is None:
            return None
        onl_des = element.find("onlineDescription", nmap)
        if onl_des is not None:
            online_des = onl_des.text
        else:
            online_des = None
        url_elem = element.find("url", nmap)
        url_func = None
        if url_elem is not None:
            url = url_elem.text
            got_url_func = url_elem.get("function", "download")
            for candid_func in EMLOnline.FunctionType:
                if got_url_func.lower() == candid_func.name.lower():
                    url_func = candid_func
                    break
        else:
            url = None
        conn_elem = element.find("connection", nmap)
        if conn_elem is not None:
            conn = EMLOnline.Connection.parse(conn_elem, nmap)
        else:
            conn = None
        des_elem = element.find("connectionDefinition", nmap)
        if des_elem is not None:
            conn_des = EMLOnline.ConnectionDefinition.parse(des_elem, nmap)
        else:
            conn_des = None
        online = EMLOnline(
            online_description=online_des,
            url=url,
            url_function=url_func,
            connection=conn,
            connection_definition=conn_des
        )
        online.__namespace__ = nmap
        return online

    def to_element(self) -> et.Element:
        """
        Generate an XML element from an Online instance.

        Returns
        -------
        lxml.etree.Element
            XML element with EMLOnline instance information.
        """
        online_elem = super().to_element()
        if self.online_description is not None:
            desc_elem = self.object_to_element("onlineDescription")
            desc_elem.text = self.online_description
            online_elem.append(desc_elem)
        if self.url is not None:
            url_elem = self.object_to_element("url")
            url_elem.set("function", self.url_function.name.lower())
            url_elem.text = self.url
            online_elem.append(url_elem)
        if self.connection is not None:
            online_elem.append(self.connection.to_element())
        if self.connection_definition is not None:
            online_elem.append(self.connection_definition.to_element())
        return online_elem



from __future__ import annotations

import re
from typing import Dict, List, Union, Tuple

from lxml import etree as et
import datetime as dt

from xml_common import XMLObject
from eml.types import EMLObject, Scope
from xml_common.utils.type_functions import unformat_datetime, format_datetime


class TemporalCoverage(EMLObject):
    """
    Temporal coverage information.

    Parameters
    ----------
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    single_datetime : List[date or datetime or AlternativeTimeScale], optional
        Means of encoding a single date and time.
    range_datetime: Tuple[date or datetime or AlternativeTimeScale], optional
        Means of encoding a range of dates and times.
    """
    PRINCIPAL_TAG = "temporalCoverage"

    class AlternativeTimeScale(XMLObject):
        """
        A name, code, or date describing an event or period in an alternative timescale.

        Parameters
        ----------
        name : str
            Name of a recognized alternative timescale.
        age_estimate : str
            Either an absolute date or a relative age name describing an event or period in an alternative timescale.
        age_uncertainty : str, optional
            The error estimate for the alternative timescale.
        age_explanation : str, optional
            The name and/or description of the method used to calculate the timescale age estimate.
        citation : List[EMLCitation], optional
            Citation for works providing detailed information about any element of the timescale age.
        """
        PRINCIPAL_TAG = "alternativeTimeScale"

        def __init__(
                self, name: str,
                age_estimate: str,
                age_uncertainty: str = None,
                age_explanation: str = None,
                citation: List = None
        ) -> None:
            super().__init__()
            self.__name__ = name
            self.__estimate__ = age_estimate
            self.__uncertainty__ = age_uncertainty
            self.__explanation__ = age_explanation
            self.__citation__ = list()
            if citation is not None:
                self.__citation__.extend(citation)
            return

        @property
        def name(self) -> str:
            """str: Name of a recognized alternative time scale."""
            return self.__name__

        @property
        def age_estimate(self) -> str:
            """
            str: Age name describing an event or period in an alternative timescale.
            """
            return self.__estimate__

        @property
        def age_uncertainty(self) -> str:
            """str: The error estimate for the alternative timescale."""
            return self.__uncertainty__

        @property
        def age_explanation(self) -> str:
            """str: The name and/or description of the method used to calculate the timescale age estimate."""
            return self.__explanation__

        @property
        def citation(self) -> List:
            """
            List[EMLCitation]: Citation for works providing detailed information about any element of the timescale age.
            """
            return self.__citation__

        @classmethod
        def parse(cls, element: et.Element, nmap: Dict) -> TemporalCoverage.AlternativeTimeScale | None:
            """
            Generate an Alternative TimeScale object from an XML element.

            Parameters
            ----------
            element : lxml.etree.Element
                XML element instance to parse.
            nmap : Dict
                Namespace.

            Returns
            -------
            AlternativeTimeScale
                Instance with the information of the XML
            """
            if element is None:
                return None
            name_elem = element.find("timeScaleName", nmap)
            estimate_elem = element.find("timeScaleAgeEstimate", nmap)
            uncertainty_elem = element.find("timeScaleAgeUncertainty", nmap)
            uncertainty = None
            if uncertainty_elem is not None:
                uncertainty = uncertainty_elem.text if uncertainty_elem.text is not None else ""
            explanation_elem = element.find("timeScaleAgeExplanation", nmap)
            explanation = None
            if explanation_elem is not None:
                explanation = explanation_elem.text if explanation_elem.text is not None else ""
            citations = list()
            from eml.resources import EMLCitation
            for cite in element.findall("timeScaleCitation", nmap):
                citations.append(EMLCitation.parse(cite, nmap))
            alter_time_scale = TemporalCoverage.AlternativeTimeScale(
                name=name_elem.text if name_elem is not None else "",
                age_estimate=estimate_elem.text if estimate_elem is not None else "",
                age_uncertainty=uncertainty,
                age_explanation=explanation,
                citation=citations
            )
            alter_time_scale.__namespace__ = nmap
            return alter_time_scale

        def to_element(self) -> et.Element:
            """
            Generate an XML element from the information of this instance.

            Returns
            -------
            lxml.etree.Element
                An XML element
            """
            element = super().to_element()
            name_elem = self.object_to_element("timeScaleName")
            name_elem.text = self.name
            element.append(name_elem)
            age_estimate_elem = self.object_to_element("timeScaleAgeEstimate")
            age_estimate_elem.text = self.age_estimate
            element.append(age_estimate_elem)
            if self.age_uncertainty is not None:
                age_uncertainty_elem = self.object_to_element("timeScaleAgeUncertainty")
                age_uncertainty_elem.text = self.age_uncertainty
                element.append(age_uncertainty_elem)
            if self.age_explanation is not None:
                age_explanation_elem = self.object_to_element("timeScaleAgeExplanation")
                age_explanation_elem.text = self.age_explanation
                element.append(age_explanation_elem)
            for citation in self.citation:
                citation_elem = citation.to_element()
                citation_elem.tag = "timeScaleCitation"
                element.append(citation_elem)
            return element

    def __init__(
            self, _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
            referencing: bool = False,
            references_system: str = None,
            single_datetime: List[Union[dt.date, dt.datetime, TemporalCoverage.AlternativeTimeScale]] = None,
            range_datetime: Tuple[
                Union[dt.date, dt.datetime, TemporalCoverage.AlternativeTimeScale],
                Union[dt.date, dt.datetime, TemporalCoverage.AlternativeTimeScale]
            ] = None,
    ) -> None:
        super().__init__(_id, scope, system, referencing, references_system)
        if (
                not self.referencing and
                (single_datetime is None or len(single_datetime) == 0) and
                range_datetime is None
        ):
            raise ValueError("Must provide at least one single datetime or a range")
        self.__range__ = range_datetime
        self.__single__ = list()
        if single_datetime is not None:
            self.__single__.extend(single_datetime)
        return

    @property
    def single_datetime(self) -> List[Union[dt.date, dt.datetime, TemporalCoverage.AlternativeTimeScale]]:
        """
        List[date or datetime or AlternativeTimeScale]: Means of encoding a single date and time.
        """
        return self.__single__

    @property
    def range_datetime(self) -> Tuple[
        Union[dt.date, dt.datetime, TemporalCoverage.AlternativeTimeScale],
        Union[dt.date, dt.datetime, TemporalCoverage.AlternativeTimeScale]
    ]:
        """
        Tuple[date or datetime or AlternativeTimeScale]: Means of encoding a range of dates and times.
        """
        return self.__range__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> TemporalCoverage:
        """
        Generate a Time Coverage referencing another Time Coverage.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        TemporalCoverage
            Object parsed that reference another.
        """
        references = element.find("references", nmap)
        return TemporalCoverage(
            _id=references.text if references.text is not None else "",
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None)
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> TemporalCoverage:
        """
        Generate a TemporalCoverage that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        TemporalCoverage
            Object parsed.
        """
        single_datetime = list()
        for single_elem in element.findall("singleDateTime", nmap):
            single_datetime.append(cls.xml_to_datetime(single_elem, nmap))
        range_datetime = None
        range_elem = element.find("rangeOfDates", nmap)
        if range_elem is not None:
            begin_elem = range_elem.find("beginDate", nmap)
            end_elem = range_elem.find("endDate", nmap)
            range_datetime = cls.xml_to_datetime(begin_elem, nmap), cls.xml_to_datetime(end_elem, nmap)
        return TemporalCoverage(
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=False,
            single_datetime=single_datetime,
            range_datetime=range_datetime,
        )

    def to_element(self) -> et.Element:
        """
        Generate an XML element with the information of the TemporalCoverage object.

        Returns
        -------
        lxml.etree.Element
            AN XML element.
        """
        element = super().to_element()
        element = self._to_element_(element)
        for single in self.single_datetime:
            single_elem = self.datetime_to_xml(single)
            single_elem.tag = "singleDateTime"
            element.append(single_elem)
        if self.range_datetime is not None:
            range_elem = self.object_to_element("rangeOfDates")
            begin_elem = self.datetime_to_xml(self.range_datetime[0])
            begin_elem.tag = "beginDate"
            range_elem.append(begin_elem)
            end_elem = self.datetime_to_xml(self.range_datetime[1])
            end_elem.tag = "endDate"
            range_elem.append(end_elem)
            element.append(range_elem)
        return element

    @staticmethod
    def xml_to_datetime(element: et.Element, nmap: Dict) -> Union[dt.date, dt.datetime, TemporalCoverage.AlternativeTimeScale]:
        """
        Convert an XML element instance to datetime object in the correspondant format.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        datetime, date or AlternativeTimeScale
            Datetime in the corresponding format.
        """
        alternative_elem = element.find("alternativeTimeScale")
        if alternative_elem is not None:
            return TemporalCoverage.AlternativeTimeScale.parse(alternative_elem, nmap)
        time_elem = element.find("time")
        calendar_elem = element.find("calendarDate")
        calendar_date = format_datetime(calendar_elem.text).date()
        if time_elem is not None:
            try:
                output_datetime = dt.datetime.strptime(
                    f"{calendar_date.strftime('%Y-%m-%d')} {time_elem.text}",
                    "%Y-%m-%d %H:%M:%SZ"
                )
                output_datetime = output_datetime.replace(tzinfo=dt.timezone.utc)
                return output_datetime
            except ValueError:
                return dt.datetime.strptime(
                    f"{calendar_date.strftime('%Y-%m-%d')} {time_elem.text}",
                    "%Y-%m-%d %H:%M:%S%z"
                )
        else:
            return calendar_date

    @staticmethod
    def datetime_to_xml(datetime: Union[dt.date, dt.datetime, TemporalCoverage.AlternativeTimeScale]) -> et.Element:
        """
        Convert a datetime to an XML element instance.

        Parameters
        ----------
        datetime : datetime, date or AlternativeTimeScale
            XML element to parse.

        Returns
        -------
        lxml.etree.Element
            XML element instance.

        Raises
        ------
        TypeError
            Not a valid datetime.
        """
        element = et.Element("placeholderDate")
        if isinstance(datetime, dt.datetime):
            string_time = unformat_datetime(datetime)
            string_time = re.sub(r"([+-])(\d{2})(\d{2})", r"\1\2:\3", string_time)
            date_text, time_text = string_time.split("T")
            calendar_elem = et.Element("calendarDate")
            calendar_elem.text = date_text
            element.append(calendar_elem)
            time_elem = et.Element("time")
            time_elem.text = time_text
            element.append(time_elem)
        elif isinstance(datetime, dt.date):
            calendar_elem = et.Element("calendarDate")
            date_text = unformat_datetime(dt.datetime.combine(datetime, dt.time()))
            calendar_elem.text = date_text
            element.append(calendar_elem)
        elif isinstance(datetime, TemporalCoverage.AlternativeTimeScale):
            element.append(datetime.to_element())
        else:
            raise TypeError("Not a valid datetime")
        return element



from __future__ import annotations

from lxml import etree as et

from enum import Enum


class EMLResource(Enum):
    """
    Type of the resources described by the EML file.
    """
    DATASET = 0
    CITATION = 1
    SOFTWARE = 2
    PROTOCOL = 3

    @classmethod
    def get_resource_type(cls, element: et.Element) -> EMLResource:
        """
        Get the resource type from the tag of one element inside another XML element.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element instance.

        Returns
        -------
        EMLResource
            Type of resource.

        Raises
        ------
        ValueError
            If resource type is not a valid type.
        """
        for resource in EMLResource:
            if element.find(resource.name.lower()) is not None:
                return resource
        raise ValueError(f"""One of the following must be present: {', '.join(
            [eml_resource.name.lower() for eml_resource in EMLResource]
        )}""")



from eml.resources.distribution.online import EMLOnline
from eml.resources.distribution.offline import EMLOffline
from eml.resources.distribution.inline import EMLInline
from eml.resources.distribution.distribution import EMLDistribution



import unittest

from lxml import etree as et

from dwca.terms import EventID
from test_xml.test_xml import TestXML


class TestStringField(TestXML):
    def test_parse(self):
        event_id_xml = """<field index="2" term="http://rs.tdwg.org/dwc/terms/eventID"/>"""
        event_id = EventID.from_string(event_id_xml)
        self.assertEqual(
            "http://rs.tdwg.org/dwc/terms/eventID",
            event_id.uri,
            "Wrong URI"
        )
        self.assertEqual(2, event_id.index, "Wrong index")
        self.assertEqualTree(et.fromstring(event_id_xml), event_id.to_element(), "Wrong generated element.")
        pass

    def test_format(self):
        event_id = EventID(2)
        value = "an_example_id"
        self.assertEqual(
            value, event_id.format(value), "Error formatting string"
        )


if __name__ == '__main__':
    unittest.main()



import unittest

from lxml import etree as et

from eml.resources import EMLSoftware
from test_xml.test_xml import TestXML


class TestEMLSoftware(TestXML):
    DEFAULT_TAGS = {
        "scope": "document"
    }

    def test_parse_referencing(self):
        text_xml = """
<software>
    <references>software-id</references>
</software>
        """
        software = EMLSoftware.from_string(text_xml)
        self.assertTrue(software.referencing, "Referencing parse incorrectly")
        self.assertEqual("software-id", software.id, "Id parse incorrectly")
        self.assertEqualTree(et.fromstring(text_xml), software.to_element(), "Error on to element")


if __name__ == '__main__':
    unittest.main()



import datetime
import datetime as dt
import unittest

from lxml import etree as et

from dwca.terms import ChronometricAgeDeterminedDate
from test_xml.test_xml import TestXML


class TestDateTimeField(TestXML):
    def test_parse(self):
        chron_determined_xml = """<field index="3" term="http://rs.tdwg.org/chrono/terms/chronometricAgeDeterminedDate"/>"""
        chron_determined = ChronometricAgeDeterminedDate.from_string(chron_determined_xml)
        self.assertEqual(
            "http://rs.tdwg.org/chrono/terms/chronometricAgeDeterminedDate",
            chron_determined.uri,
            "Wrong URI"
        )
        self.assertEqual(3, chron_determined.index, "Wrong index")
        self.assertEqualTree(
            et.fromstring(chron_determined_xml),
            chron_determined.to_element(),
            "Wrong generated element."
        )
        pass

    def test_format(self):
        chron_determined = ChronometricAgeDeterminedDate(3)
        value = "1963-03-08T14:07-0600"
        self.assertEqual(
            dt.datetime(year=1963, month=3, day=8, hour=14, minute=7, tzinfo=dt.timezone(dt.timedelta(hours=-6))),
            chron_determined.format(value),
            f"Error formatting datetime ({value})"
        )
        value = "2009-02-20T08:40Z"
        self.assertEqual(
            dt.datetime(year=2009, month=2, day=20, hour=8, minute=40, tzinfo=datetime.timezone.utc),
            chron_determined.format(value),
            f"Error formatting datetime ({value})"
        )
        value = "2018-08-29T15:19"
        self.assertEqual(
            dt.datetime(year=2018, month=8, day=29, hour=15, minute=19),
            chron_determined.format(value),
            f"Error formatting datetime ({value})"
        )
        value = "1809-02-12"
        self.assertEqual(
            dt.datetime(year=1809, month=2, day=12),
            chron_determined.format(value),
            f"Error formatting datetime ({value})"
        )
        value = "1906-06"
        self.assertEqual(
            dt.datetime(year=1906, month=6, day=1),
            chron_determined.format(value),
            f"Error formatting datetime ({value})"
        )
        value = "1971"
        self.assertEqual(
            dt.datetime(year=1971, month=1, day=1),
            chron_determined.format(value),
            f"Error formatting datetime ({value})"
        )
        value = "2007-03-01T13:00:00Z/2008-05-11T15:30:00Z"
        self.assertRaisesRegex(
            ValueError, "does not match any", chron_determined.format, value
        )


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import Dict

from lxml import etree as et

from test_xml.test_xml import TestXML
from xml_common import XMLObject


class TestSimpleXML(TestXML):

    class SimpleXML(XMLObject):
        PRINCIPAL_TAG = "simple"

        def __init__(self, an_attribute: str, another_attribute: str) -> None:
            super().__init__()
            self.__an_attribute__ = an_attribute
            self.__another_attribute__ = another_attribute
            return

        @classmethod
        def parse(cls, element: et.Element, nmap: Dict) -> XMLObject | None:
            an_attribute = element.find("anAttribute")
            another_attribute = element.find("anotherAttribute")
            simple = TestSimpleXML.SimpleXML(
                an_attribute.text, another_attribute.text
            )
            simple.__namespace__ = nmap
            return simple

        def to_element(self) -> et.Element:
            simple_element = et.Element(self.PRINCIPAL_TAG)
            an_attribute = et.Element("anAttribute")
            an_attribute.text = self.__an_attribute__
            simple_element.append(an_attribute)
            another_attribute = et.Element("anotherAttribute")
            another_attribute.text = self.__another_attribute__
            simple_element.append(another_attribute)
            return simple_element

    def test_parse(self) -> None:
        simple_xml = """
<simple>
    <anAttribute>123</anAttribute>
    <anotherAttribute>456</anotherAttribute>
</simple>
        """
        simple = TestSimpleXML.SimpleXML.from_string(simple_xml)
        self.assertEqual(
            "123", simple.__an_attribute__, "Error on parsing `an attribute`."
        )
        self.assertEqual(
            "456", simple.__another_attribute__, "Error on parsing `another attribute`."
        )
        self.assertEqualTree(et.fromstring(simple_xml), simple.to_element(), "Error on `to_element`.")

    def test_repr(self):
        simple = TestSimpleXML.SimpleXML("an_attribute", "another_attribute")
        self.assertEqual(
            "<XMLObject tag=simple>",
            repr(simple),
            "Incorrect representation"
        )
        self.assertEqual(
            "<XMLObject tag=simple>",
            f"{simple}",
            "Incorrect representation"
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import Dict

from lxml import etree as et

from eml.types import I18nString


class OrganizationName(I18nString):
    """
    The full name of the organization being described.
    """
    PRINCIPAL_TAG = "organizationName"

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> OrganizationName | None:
        """
        Parses a lxml element into an Organization Name.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Python object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        OrganizationName
            Object parsed.
        """
        if element is None:
            return None
        return OrganizationName(element.text, lang=element.get(f"{{{cls.NAMESPACES['xml']}}}lang", "eng"))

    @classmethod
    def check_principal_tag(cls, tag: str, nmap: Dict) -> None:
        """
        Checks if the tag is "organizationName"

        Parameters
        ----------
        tag : str
            Tag obtained from outside source.
        nmap : Dict
            The namespace to check the format of the principal tag.

        Raises
        ------
        AssertionError
            If the tag is not a representation of the principal tag of the object.
        """
        return super(I18nString, cls).check_principal_tag(tag, nmap)

    def set_tag(self, tag: str) -> None:
        """
        To not be used.

        Parameters
        ----------
        tag : str
            Any tag.
        """
        return

    def to_element(self) -> et.Element:
        """
        Generate a lxml.tree.Element from the organization name.

        Returns
        -------
        lxml.tree.Element
            Object in the XML Element format.
        """
        element = et.Element(self.PRINCIPAL_TAG)
        element.text = self.__value__
        element.set(f"{{{self.__namespace__['xml']}}}lang", self.__lang__.name.lower())
        return element

    def __repr__(self) -> str:
        return f"<Organization Name ({self.__value__}) [{self.language.name.lower()}]>"



import unittest

from lxml import etree as et
from dwca.terms import OutsideTerm


class TestField(unittest.TestCase):
    def test_parse(self):
        field = OutsideTerm.from_string('<field index="1" term="http://rs.tdwg.org/dwc/terms/scientificNameID"/>')
        self.assertEqual(1, field.index, "Wrong parse index")
        self.assertEqual("http://rs.tdwg.org/dwc/terms/scientificNameID", field.uri, "Wrong parse term")
        self.assertIsNone(field.default, "Wrong parse default value")
        self.assertIsNone(field.vocabulary, "Wrong parse vocabulary value")

    def test_parse_invalid(self):
        self.assertRaises(
            ValueError,
            OutsideTerm.from_string,
            '<field index="OK" term="http://rs.tdwg.org/dwc/terms/scientificNameID"/>'
        )
        self.assertRaises(
            AssertionError,
            OutsideTerm.from_string,
            '<archive index="0" term="Valid Term"/>'
        )
        self.assertRaises(
            TypeError,
            OutsideTerm.from_string,
            '<field index="2"/>'
        )

    def test_xml(self):
        field = OutsideTerm(index=1, uri="http://rs.tdwg.org/dwc/terms/scientificNameID")
        expected = et.Element("field")
        expected.set("term", "http://rs.tdwg.org/dwc/terms/scientificNameID")
        expected.set("index", "1")
        actual = et.fromstring(field.to_xml())
        self.assertEqual(et.tostring(expected), et.tostring(actual), "Wrong xml generated")


if __name__ == '__main__':
    unittest.main()



import unittest

from lxml import etree as et

from dwca.terms import OutsideTerm, TaxonID, TaxonRank
from test_xml.test_xml import TestXML


class TestTaxon(TestXML):
    def test_taxon_id(self):
        term_xml = """
<field index="0" term="http://rs.tdwg.org/dwc/terms/taxonID"/>
        """
        term = TaxonID.from_string(term_xml)
        self.assertEqual("taxonID", term.name, "Name parsed incorrectly.")
        self.assertEqual("VARCHAR", term.sql_type, "Wrong SQL type.")
        self.assertEqual(0, term.index, "Index wrongly parsed.")
        self.assertIsNone(term.default, "Default value from nowhere.")
        self.assertIsNone(term.vocabulary, "Vocabulary from nowhere.")
        self.assertEqual("http://rs.tdwg.org/dwc/terms/taxonID", term.uri, "URI wrongly parsed")
        self.assertEqualTree(et.fromstring(term_xml), term.to_element(), "Error on to element")

    def test_taxon_rank(self):
        term_xml = """
<field
    index="0"
    term="http://rs.tdwg.org/dwc/terms/taxonRank"
    default="species"
    vocabulary="https://www.iapt-taxon.org/nomen/pages/main/art_3.html"
/>
        """
        term = TaxonRank.from_string(term_xml)
        self.assertEqual("taxonRank", term.name, "Name parsed incorrectly.")
        self.assertEqual("VARCHAR", term.sql_type, "Wrong SQL type.")
        self.assertEqual(0, term.index, "Index wrongly parsed")
        self.assertEqual("species", term.default, "Default wrongly parsed")
        self.assertEqual("https://www.iapt-taxon.org/nomen/pages/main/art_3.html", term.vocabulary, "Vocabulary wrongly parsed")
        self.assertEqual("http://rs.tdwg.org/dwc/terms/taxonRank", term.uri, "URI wrongly parsed")
        self.assertEqualTree(et.fromstring(term_xml), term.to_element(), "Error on to element")

    def test_repr(self):
        term_xml = """
<field index="0" term="http://rs.tdwg.org/dwc/terms/taxonID"/>
        """
        term = TaxonID.from_string(term_xml)
        self.assertEqual(
            "<Field [term=http://rs.tdwg.org/dwc/terms/taxonID]>",
            repr(term),
            "Taxon ID wrongly parsed"
        )

    def test_parse_none(self):
        self.assertIsNone(
            TaxonID.parse(None, {}),
            "Parse term from nowhere"
        )

    def test_parse_no_field(self):
        self.assertRaises(
            TypeError,
            TaxonID.from_string,
            """
<field index="0"/>
            """
        )


if __name__ == '__main__':
    unittest.main()



import unittest

from lxml import etree as et

from xml_common.utils import Language
from test_xml.test_xml import TestXML
from eml.types import OrganizationName


class TestOrganizationName(TestXML):
    def test_organization_name(self):
        variable = OrganizationName("Institute of Ecology and Biodiversity")
        self.assertEqual("Institute of Ecology and Biodiversity", variable, "Not equal to string")
        self.assertEqual("Institute of Ecology and Biodiversity", str(variable), "Not equal to string")
        self.assertEqual(
            "<Organization Name (Institute of Ecology and Biodiversity) [eng]>",
            repr(variable),
            "Not equal to string"
        )
        self.assertEqual(Language.ENG, variable.language, "Language set incorrectly")

    def test_organization_name_spanish(self):
        variable = OrganizationName("Instituto de Ecología y Biodiversidad", lang=Language.ESP)
        self.assertEqual("Instituto de Ecología y Biodiversidad", variable, "Not equal to string")
        self.assertEqual("Instituto de Ecología y Biodiversidad", str(variable), "Not equal to string")
        self.assertEqual(
            "<Organization Name (Instituto de Ecología y Biodiversidad) [esp]>",
            repr(variable),
            "Not equal to string"
        )
        self.assertEqual(Language.ESP, variable.language, "Language set incorrectly")

    def test_parse(self):
        self.assertIsNone(OrganizationName.parse(None, {}), "Parsing something")
        example_xml = """
<organizationName
    xml:lang="eng"
>Institute of Ecology and Biodiversity</organizationName>
        """
        variable = OrganizationName.from_string(example_xml)
        self.assertEqual("Institute of Ecology and Biodiversity", variable, "Not equal to string")
        self.assertEqual(Language.ENG, variable.language, "Language set incorrectly")

    def test_wrong_language_type(self):
        self.assertRaises(
            TypeError, OrganizationName, "My variable", 20
        )

    def test_to_element(self):
        variable = OrganizationName("Instituto de Ecología y Biodiversidad", lang=Language.ESP)
        expected = et.Element("organizationName")
        expected.text = "Instituto de Ecología y Biodiversidad"
        expected.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        self.assertEqualTree(expected, variable.to_element(), "Error on element")
        variable.set_tag("AnInvalidTag")
        self.assertEqualTree(expected, variable.to_element(), "Error on element")


if __name__ == '__main__':
    unittest.main()



import datetime as dt
from typing import Any, List, Union

from datetime_interval import Interval

from dwca.terms import Field


class MeasurementID(Field):
    """
    An identifier for the MeasurementOrFact.

    Information pertaining to measurements,
    facts, characteristics, or assertions.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ParentMeasurementID(Field):
    """
    An identifier for a broader MeasurementOrFact that groups this and potentially other MeasurementOrFacts.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/parentMeasurementID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MeasurementType(Field):
    """
    The nature of the measurement, fact, characteristic, or assertion.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementType"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class MeasurementValue(Field):
    """
    The value of the measurement, fact, characteristic, or assertion.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementValue"
    TYPE = Any

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class MeasurementAccuracy(Field):
    """
    The description of the potential error associated with the MeasurementValue.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementAccuracy"
    TYPE = Any

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MeasurementUnit(Field):
    """
    The units associated with the MeasurementValue.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementUnit"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MeasurementDeterminedBy(Field):
    """
    A list of names of people, groups, or organizations who determined the value of the MeasurementOrFact.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementDeterminedBy"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MeasurementDeterminedDate(Field):
    """
    The date on which the MeasurementOrFact was made.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementDeterminedDate"
    TYPE = Union[dt.datetime, Interval]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MeasurementMethod(Field):
    """
    A description of or reference to the method or protocol used to determine the MeasurementOrFact.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementMethod"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MeasurementRemarks(Field):
    """
    Comments or notes accompanying the MeasurementOrFact.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/measurementRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



from unittest import main
from lxml import etree as et
from pygments.lexers import q

from eml.resources import EMLCitation, EMLProtocol
from eml.types import ProcedureStep, EMLTextType
from test_xml.test_xml import TestXML


class TestProcedureStep(TestXML):
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
        'scope': 'document',
    }

    def test_parse_none(self):
        self.assertIsNone(ProcedureStep.parse(None, {}), "Procedure Step from nowhere.")

    def test_parse_simple(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
</methodStep>
        """
        procedure = ProcedureStep.from_string(xml_text)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            procedure.description.paragraphs[0],
            "Error parsing description."
        )
        self.assertIsNone(procedure.citation, "Citation from nowhere.")
        self.assertIsNone(procedure.protocol, "Protocol from nowhere.")
        self.assertEqual(0, len(procedure.instrumentation), "Instrumentation from nowhere.")
        self.assertEqual(0, len(procedure.software), "Software from nowhere.")
        self.assertEqual(0, len(procedure.sub_step), "Sub step from nowhere.")
        procedure.set_tag("methodStep")
        self.assertEqualTree(et.fromstring(xml_text), procedure.to_element(), "Error on element conversion.")

    def test_citation_protocol_error(self):
        citation = EMLCitation(_id="cite-id", referencing=True)
        protocol = EMLProtocol(_id="protocol-id", referencing=True)
        with self.assertRaisesRegex(TypeError, "allow only one"):
            ProcedureStep(description=EMLTextType(paragraphs=["Lorem ipsum"]), citation=citation, protocol=protocol)

    def test_parse_error(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
    <citation>
        <references>cite-id</references>
    </citation>
    <protocol>
        <references>protocol-id</references>
    </protocol>
</methodStep>
        """
        with self.assertRaisesRegex(TypeError, "allow only one"):
            ProcedureStep.from_string(xml_text)

    def test_parse_citation(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
    <citation>
        <references>cite-id</references>
    </citation>
</methodStep>
        """
        procedure = ProcedureStep.from_string(xml_text)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            procedure.description.paragraphs[0],
            "Error parsing description."
        )
        self.assertEqual("cite-id", procedure.citation.id, "Error parsing citation.")
        self.assertIsNone(procedure.protocol, "Protocol from nowhere.")
        self.assertEqual(0, len(procedure.instrumentation), "Instrumentation from nowhere.")
        self.assertEqual(0, len(procedure.software), "Software from nowhere.")
        self.assertEqual(0, len(procedure.sub_step), "Sub step from nowhere.")
        procedure.set_tag("methodStep")
        self.assertEqualTree(et.fromstring(xml_text), procedure.to_element(), "Error on element conversion.")

    def test_parse_protocol(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
    <protocol>
        <references>protocol-id</references>
    </protocol>
</methodStep>
        """
        procedure = ProcedureStep.from_string(xml_text)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            procedure.description.paragraphs[0],
            "Error parsing description."
        )
        self.assertIsNone(procedure.citation, "Citation from nowhere.")
        self.assertEqual("protocol-id", procedure.protocol.id, "Error parsing protocol.")
        self.assertEqual(0, len(procedure.instrumentation), "Instrumentation from nowhere.")
        self.assertEqual(0, len(procedure.software), "Software from nowhere.")
        self.assertEqual(0, len(procedure.sub_step), "Sub step from nowhere.")
        procedure.set_tag("methodStep")
        self.assertEqualTree(et.fromstring(xml_text), procedure.to_element(), "Error on element conversion.")

    def test_parse_single_instrumentation(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
    <instrumentation>LACHAT analyzer, model XYX</instrumentation>
</methodStep>
        """
        procedure = ProcedureStep.from_string(xml_text)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            procedure.description.paragraphs[0],
            "Error parsing description."
        )
        self.assertIsNone(procedure.citation, "Citation from nowhere.")
        self.assertIsNone(procedure.protocol, "Protocol from nowhere.")
        self.assertEqual(1, len(procedure.instrumentation), "Error parsing single instrumentation.")
        self.assertEqual("LACHAT analyzer, model XYX", procedure.instrumentation[0], "Error parsing instrumentation.")
        self.assertEqual(0, len(procedure.software), "Software from nowhere.")
        self.assertEqual(0, len(procedure.sub_step), "Sub step from nowhere.")
        procedure.set_tag("methodStep")
        self.assertEqualTree(et.fromstring(xml_text), procedure.to_element(), "Error on element conversion.")

    def test_parse_instrumentation(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
    <instrumentation>LACHAT analyzer, model XYX</instrumentation>
    <instrumentation>Microscope, model ABBA</instrumentation>
</methodStep>
        """
        procedure = ProcedureStep.from_string(xml_text)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            procedure.description.paragraphs[0],
            "Error parsing description."
        )
        self.assertIsNone(procedure.citation, "Citation from nowhere.")
        self.assertIsNone(procedure.protocol, "Protocol from nowhere.")
        self.assertEqual(2, len(procedure.instrumentation), "Error parsing instrumentations.")
        self.assertEqual("Microscope, model ABBA", procedure.instrumentation[1], "Error parsing second instrumentation.")
        self.assertEqual(0, len(procedure.software), "Software from nowhere.")
        self.assertEqual(0, len(procedure.sub_step), "Sub step from nowhere.")
        procedure.set_tag("methodStep")
        self.assertEqualTree(et.fromstring(xml_text), procedure.to_element(), "Error on element conversion.")

    def test_parse_software(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
    <software>
        <references>software-id</references>
    </software>
</methodStep>
        """
        procedure = ProcedureStep.from_string(xml_text)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            procedure.description.paragraphs[0],
            "Error parsing description."
        )
        self.assertIsNone(procedure.citation, "Citation from nowhere.")
        self.assertIsNone(procedure.protocol, "Protocol from nowhere.")
        self.assertEqual(0, len(procedure.instrumentation), "Instrumentation from nowhere.")
        self.assertEqual("software-id", procedure.software[0].id, "Error parsing software.")
        self.assertEqual(0, len(procedure.sub_step), "Sub step from nowhere.")
        procedure.set_tag("methodStep")
        self.assertEqualTree(et.fromstring(xml_text), procedure.to_element(), "Error on element conversion.")

    def test_parse_sub_step(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
    <subStep>
        <description>
            <para>Nested sub step</para>
        </description>
    </subStep>
</methodStep>
        """
        procedure = ProcedureStep.from_string(xml_text)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            procedure.description.paragraphs[0],
            "Error parsing description."
        )
        self.assertIsNone(procedure.citation, "Citation from nowhere.")
        self.assertIsNone(procedure.protocol, "Protocol from nowhere.")
        self.assertEqual(0, len(procedure.instrumentation), "Instrumentation from nowhere.")
        self.assertEqual(0, len(procedure.software), "Software from nowhere.")
        self.assertEqual(1, len(procedure.sub_step), "Error parsing sub step.")
        self.assertEqual(
            "Nested sub step",
            procedure.sub_step[0].description.paragraphs[0],
            "Error parsing sub step description."
        )
        self.assertIsNone(procedure.sub_step[0].citation, "Citation from nowhere.")
        self.assertIsNone(procedure.sub_step[0].protocol, "Protocol from nowhere.")
        self.assertEqual(0, len(procedure.sub_step[0].instrumentation), "Instrumentation from nowhere.")
        self.assertEqual(0, len(procedure.sub_step[0].software), "Software from nowhere.")
        self.assertEqual(0, len(procedure.sub_step[0].sub_step), "Sub sub step from nowhere.")
        procedure.set_tag("methodStep")
        self.assertEqualTree(et.fromstring(xml_text), procedure.to_element(), "Error on element conversion.")

    def test_parse_complex_sub_step(self):
        xml_text = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
    <subStep>
        <description>
            <para>Nested sub step</para>
        </description>
        <subStep>
            <description>
                <para>First second nested sub step</para>
            </description>
        </subStep>
        <subStep>
            <description>
                <para>Second second nested sub step</para>
            </description>
        </subStep>
    </subStep>
</methodStep>
        """
        procedure = ProcedureStep.from_string(xml_text)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            procedure.description.paragraphs[0],
            "Error parsing description."
        )
        self.assertIsNone(procedure.citation, "Citation from nowhere.")
        self.assertIsNone(procedure.protocol, "Protocol from nowhere.")
        self.assertEqual(0, len(procedure.instrumentation), "Instrumentation from nowhere.")
        self.assertEqual(0, len(procedure.software), "Software from nowhere.")
        self.assertEqual(1, len(procedure.sub_step), "Error parsing sub step.")
        self.assertEqual(
            "Nested sub step",
            procedure.sub_step[0].description.paragraphs[0],
            "Error parsing sub step description."
        )
        self.assertIsNone(procedure.sub_step[0].citation, "Citation from nowhere.")
        self.assertIsNone(procedure.sub_step[0].protocol, "Protocol from nowhere.")
        self.assertEqual(0, len(procedure.sub_step[0].instrumentation), "Instrumentation from nowhere.")
        self.assertEqual(0, len(procedure.sub_step[0].software), "Software from nowhere.")
        self.assertEqual(2, len(procedure.sub_step[0].sub_step), "Error parsing nested sub step.")
        self.assertEqual(
            "Second second nested sub step",
            procedure.sub_step[0].sub_step[1].description.paragraphs[0],
            "Error parsing nested sub step description."
        )
        procedure.set_tag("methodStep")
        self.assertEqualTree(et.fromstring(xml_text), procedure.to_element(), "Error on element conversion.")


if __name__ == '__main__':
    main()



from __future__ import annotations

from enum import Enum
from typing import Dict, List, Union

from lxml import etree as et

from xml_common import XMLObject
from eml.types import I18nString


class KeywordType(Enum):
    """
    The type of each keyword.

    Attributes
    ----------
    PLACE : int
        Keywords pertaining to a spatial location.
    STRATUM : int
        Keywords pertaining to a vertical stratum.
    TEMPORAL : int
        Keywords pertaining to temporal data.
    THEME : int
        Keywords pertaining to thematic subject.
    TAXONOMIC : int
        Keywords pertaining to taxon information.
    """
    NULL = 0
    PLACE = 1
    STRATUM = 2
    TEMPORAL = 4
    THEME = 5
    TAXONOMIC = 6


class EMLKeywordSet(XMLObject):
    """
    Keyword information that describes the resource.

    Parameters
    ----------
    keywords : List[I18nString, str]
        A list of keywords that describes the resource.
    thesaurus : str, optional
        The name of a thesaurus from which the keyword is derived.
    keywords_type : List[KeywordType], optional
        A list of the type of each keyword.
    """
    PRINCIPAL_TAG = "keywordSet"

    def __init__(
            self, keywords: List[Union[I18nString, str]],
            thesaurus: str = None,
            keywords_type: List[KeywordType] = None
    ) -> None:
        super().__init__()
        self.__keywords__: List[I18nString] = list()
        if len(keywords) <= 0:
            raise ValueError("KeywordSet needs at least one keyword")
        for keyword in keywords:
            self.__keywords__.append(I18nString(keyword))
        self.__thesaurus__ = thesaurus
        self.__keywords_type__: List[KeywordType] = list()
        if keywords_type is not None:
            self.__keywords_type__.extend(keywords_type)
        missing_types = len(self.__keywords__) - len(self.__keywords_type__)
        if missing_types:
            self.__keywords_type__.extend(
                [KeywordType.NULL] * missing_types
            )
        return

    @property
    def keywords(self) -> List[I18nString]:
        """List[I18nString]: A list of keywords that describes the resource."""
        return self.__keywords__

    @property
    def thesaurus(self) -> str:
        """str: The name of a thesaurus from which the keyword is derived."""
        return self.__thesaurus__

    @property
    def keywords_type(self) -> List[KeywordType]:
        """List[KeywordType]: A list of the type of each keyword."""
        return self.__keywords_type__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLKeywordSet | None:
        """
        Generate an EMLKeywordSet from an XML element.
        
        Parameters
        ----------
        element : lxml.etree.Element
            XML Element to be parsed.
        nmap : Dict
            Namespace of the XML element.

        Returns
        -------
        EMLKeyword
            Parsed object
        """
        if element is None:
            return None
        keywords = list()
        keyword_types = list()
        for keyword in element.findall("keyword", nmap):
            keywords.append(I18nString.parse(keyword, nmap))
            key_type_elem = keyword.get("keywordType", None)
            if key_type_elem is None:
                keyword_types.append(KeywordType.NULL)
            else:
                found = False
                for key_type in KeywordType:
                    if key_type.name.lower() == key_type_elem.lower():
                        keyword_types.append(key_type)
                        found = True
                        break
                if not found:
                    keyword_types.append(KeywordType.NULL)
        thesaurus_elem = element.find("keywordThesaurus", nmap)
        if thesaurus_elem is not None:
            thesaurus = thesaurus_elem.text if thesaurus_elem.text is not None else ""
        else:
            thesaurus = None
        keyword_set = EMLKeywordSet(
            keywords=keywords,
            thesaurus=thesaurus,
            keywords_type=keyword_types
        )
        keyword_set.__namespace__ = nmap
        return keyword_set

    def to_element(self) -> et.Element:
        """
        Get an XML Element from the EMLKeywordSet object

        Returns
        -------
        lxml.etree.Element
            XML Element with the information of the EMLAbstract instance
        """
        keyword_set_elem = super().to_element()
        for keyword, keyword_type in zip(self.keywords, self.keywords_type):
            keyword.set_tag("keyword")
            key_elem = keyword.to_element()
            if keyword_type != KeywordType.NULL:
                key_elem.set("keywordType", keyword_type.name.lower())
            keyword_set_elem.append(key_elem)
        if self.thesaurus is not None:
            thesaurus_elem = self.object_to_element("keywordThesaurus")
            thesaurus_elem.text = self.thesaurus
            keyword_set_elem.append(thesaurus_elem)
        return keyword_set_elem



import unittest

from lxml import etree as et

from dwca.terms import DecimalLatitude
from test_xml.test_xml import TestXML


class TestFloatField(TestXML):
    def test_parse(self):
        decimal_xml = """<field index="3" term="http://rs.tdwg.org/dwc/terms/decimalLatitude"/>"""
        decimal = DecimalLatitude.from_string(decimal_xml)
        self.assertEqual(
            "http://rs.tdwg.org/dwc/terms/decimalLatitude",
            decimal.uri,
            "Wrong URI"
        )
        self.assertEqual(3, decimal.index, "Wrong index")
        self.assertEqualTree(et.fromstring(decimal_xml), decimal.to_element(), "Wrong generated element.")
        pass

    def test_format(self):
        decimal = DecimalLatitude(3)
        value = "92.0"
        self.assertGreaterEqual(
            1e-10, abs(decimal.format(value) - 92.0), "Error formatting float"
        )
        value = "18.00005"
        self.assertGreaterEqual(
            1e-10, abs(decimal.format(value) - 18.00005), "Error formatting float"
        )
        value = "Invalid float"
        self.assertIsNone(decimal.format(value), "Error formatting invalid float")


if __name__ == '__main__':
    unittest.main()



from eml.types.no_tag_object import _NoTagObject
from eml.types.extension_string import ExtensionString
from eml.types.extension_uri import ExtensionURI
from eml.types.eml_object import EMLObject, Scope
from eml.types.i18n_string import I18nString
from eml.types.organization_name import OrganizationName
from eml.types.position_name import PositionName
from eml.types.individual_name import IndividualName
from eml.types.address import EMLAddress
from eml.types.phone import EMLPhone
from eml.types.responsible_party import ResponsibleParty, Role
from eml.types.section import EMLSection
from eml.types.text_type import EMLTextType
from eml.types.semantic_annotation import SemanticAnnotation
from eml.types.access_type import AccessType, AccessRole, AccessPermission
from eml.types.change_history import ChangeHistory
from eml.types.maintance import MaintUpFreqType, Maintenance
from eml.types.procedure_step import ProcedureStep
from eml.types.methods import Methods
from eml.types.research_project import ResearchProject



from __future__ import annotations

from math import frexp
from typing import Dict, List

from lxml import etree as et

from eml.types import EMLTextType, ChangeHistory
from xml_common import XMLObject
from xml_common.utils import CamelCaseEnum


class MaintUpFreqType(CamelCaseEnum):
    """
    Maintenance Update Frequency.

    Frequency with which changes and additions are made to the
    dataset after the initial dataset is completed.

    Attributes
    ----------
    ANNUALLY : int
        Every year.
    AS_NEEDED : int
        Maintenance as needed.
    BIANNUALLY : int
        Every two years.
    CONTINUALLY : int
        Maintenance continually.
    DAILY : int
        Every day.
    IRREGULAR : int
        Irregular maintenance.
    MONTHLY : int
        Every month.
    NOT_PLANNED : int
        Maintenance not planned.
    WEEKLY : int
        Every week.
    UNKNOWN : int
        Unknown maintenance.
    UNKOWN : int
        Unknown with typo.
    OTHER_MAINTENANCE_PERIOD : int
        Other unidentified maintenance period.
    """
    ANNUALLY = 1
    AS_NEEDED = 2
    BIANNUALLY = 3
    CONTINUALLY = 4
    DAILY = 5
    IRREGULAR = 6
    MONTHLY = 7
    NOT_PLANNED = 8
    WEEKLY = 9
    UNKNOWN = 10
    UNKOWN = 11
    OTHER_MAINTENANCE_PERIOD = 12


class Maintenance(XMLObject):
    """
    Class representing a description of the maintenance of a data resource.

    Parameters
    ----------
    description : EMLTextType
        A text description of the maintenance of this data resource.
    maintenance_update_frequency : MaintUpFreqType, optional
        Frequency with which changes and additions are made to the dataset after the initial dataset is completed.
    change_history : List[ChangeHistory], optional
        A description of changes made to the data since its release.
    """
    PRINCIPAL_TAG = "maintenance"

    def __init__(
            self, description: EMLTextType,
            maintenance_update_frequency: MaintUpFreqType = None,
            change_history: List[ChangeHistory] = None
    ):
        super().__init__()
        self.__description__ = description
        self.__muf__ = maintenance_update_frequency
        self.__c_hist__ = list()
        if change_history is not None:
            self.__c_hist__.extend(change_history)
        return

    @property
    def description(self) -> EMLTextType:
        """EMLTextType: A text description of the maintenance of this data resource."""
        return self.__description__

    @property
    def maintenance_update_frequency(self) -> MaintUpFreqType:
        """MaintUpFreqType: Frequency with which changes and additions are made to the dataset after the initial dataset is completed."""
        return self.__muf__

    @property
    def change_history(self) -> List[ChangeHistory]:
        """List[ChangeHistory]: A description of changes made to the data since its release."""
        return self.__c_hist__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> Maintenance | None:
        """
        Parses a lxml element into a Maintenance object.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Maintenance object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        Maintenance
            Object parsed.
        """
        if element is None:
            return None
        freq_elem = element.find("maintenanceUpdateFrequency", nmap)
        muf = MaintUpFreqType.get_enum(freq_elem.text) if freq_elem is not None else None
        change_history_elems = element.findall("changeHistory", nmap)
        change_history = list()
        for ch in change_history_elems:
            change_history.append(ChangeHistory.parse(ch, nmap))
        return Maintenance(
            description=EMLTextType.parse(element.find("description", nmap), nmap),
            maintenance_update_frequency=muf,
            change_history=change_history
        )


    def to_element(self) -> et.Element:
        """
        Generates a lxml.tree.Element from the Maintenance object.

        Returns
        -------
        lxml.tree.Element
            Object in the Element format.
        """
        element = super().to_element()
        self.description.set_tag("description")
        element.append(self.description.to_element())
        if self.maintenance_update_frequency is not None:
            muf_elem = self.object_to_element("maintenanceUpdateFrequency")
            muf_elem.text = self.maintenance_update_frequency.to_camel_case()
            element.append(muf_elem)
        for ch in self.change_history:
            element.append(ch.to_element())
        return element



from __future__ import annotations

from typing import Dict, List, Tuple, Any

import datetime as dt
from lxml import etree as et

from xml_common.utils import Language
from eml.base import EMLMetadata, EMLVersion
from eml.resources import EMLResource, EMLDataset, EMLCitation, EMLProtocol, EMLSoftware, Resource, EMLKeywordSet, \
    EMLLicense, EMLDistribution, EMLCoverage
from eml.types import Scope, SemanticAnnotation, AccessType, EMLObject, I18nString, ResponsibleParty, ExtensionString, \
    Role, EMLTextType


class EML(EMLObject):
    """
    Class representing an `Ecological Metadata Language <https://eml.ecoinformatics.org/>`_

    Parameters
    ----------
    package_id : str
        A globally unique identifier for the data package described by this EML that can be used to cite it elsewhere.
    system : str
        The data management system within which an identifier is in scope and therefore unique.
    resource_type : EMLResource
        Type of the resource: `dataset`, `citation`, `protocol` or `software.`
    version: EMLVersion, optional
        Version of the EML standard. Default: latest (2.2.0).
    language : Language, optional
        Language abbreviation to be used, defaults to `"eng"`
    access : AccessType, optional
        Access control rules for the entire resource, which can be overridden by access rules in distribution trees.
    additional_metadata : List[None], optional
        A flexible field for including any other relevant metadata that pertains to the resource being described.
    annotation: List[Tuple[SemanticAnnotation, str]], optional
        A list of precisely-defined semantic statements about this resource.
    """
    PRINCIPAL_TAG = "eml"
    NAMESPACE_TAG = "eml"
    """str: Namespace tag `eml`, to be replace for `{uri}tag`"""

    class Annotation:
        """
        A precisely-defined semantic statement about an element in the EML document.

        Parameters
        ----------
        semantic_annotation: SemanticAnnotation
            A semantically-precise statement about an EML element.
        references : str
            The id of the element being annotated.
        """
        def __init__(self, semantic_annotation: SemanticAnnotation, references: str) -> None:
            self.__annotation__ = semantic_annotation
            self.__ref__ = references
            return

        @property
        def annotation(self) -> SemanticAnnotation:
            """SemanticAnnotation: A semantically-precise statement about an EML element."""
            return self.__annotation__

        @property
        def references(self) -> str:
            """str: The id of the element being annotated."""
            return self.__ref__

        def __eq__(self, other: Any) -> bool:
            if isinstance(other, EML.Annotation):
                return (
                        self.annotation == other.annotation and
                        self.references == other.references
                )
            elif isinstance(other, SemanticAnnotation):
                return self.annotation == other
            else:
                return False

    def __init__(
            self,
            package_id: str,
            system: str,
            resource_type: EMLResource,
            version: EMLVersion = EMLVersion.LATEST,
            language: Language = Language.ENG,
            access: AccessType = None,
            additional_metadata: List[Any] = None,
            annotation: List[Tuple[SemanticAnnotation, str]] = None
    ) -> None:
        super().__init__(None, Scope.SYSTEM, system, False)
        self.__schema_location__ = ""
        self.__package__ = package_id
        self.__lang__ = language
        self.__access__ = None
        self.__resource_type__ = resource_type
        self.__version__ = version
        if resource_type == EMLResource.DATASET:
            self.__resource_class__ = EMLDataset
        elif resource_type == EMLResource.CITATION:
            self.__resource_class__ = EMLCitation
        elif resource_type == EMLResource.SOFTWARE:
            self.__resource_class__ = EMLSoftware
        elif resource_type == EMLResource.PROTOCOL:
            self.__resource_class__ = EMLProtocol
        else:
            raise ValueError(f"{resource_type.name.lower()} is not allowed")
        self.__resource__ = None
        self.__kwargs__ = dict()
        self.__access__ = access
        self.__additional__ = list()
        if additional_metadata is not None:
            self.__additional__.extend(additional_metadata)
        self.__annotation__ = list()
        if annotation is not None:
            self.__annotation__.extend([EML.Annotation(a[0], a[1]) for a in annotation])
        return

    @property
    def package_id(self) -> str:
        """str: A globally unique identifier for the data package described by this EML that can be used to cite it."""
        return self.__package__

    @property
    def system(self) -> str:
        """str: The data management system within which an identifier is in scope and therefore unique."""
        return self.__system__

    @property
    def scope(self) -> Scope:
        """Scope: The scope of the identifier."""
        return self.__scope__

    @property
    def language(self) -> Language:
        """Language: The language of the resource."""
        return self.__lang__

    @property
    def resource_type(self) -> EMLResource:
        """EMLResource: The type of the resource."""
        return self.__resource_type__

    @property
    def resource(self) -> Resource:
        """Resource: The resource instance."""
        if self.__resource__ is None:
            try:
                self.__resource__ = self.__resource_class__(**self.__kwargs__)
                return self.__resource__
            except Exception as _:
                raise RuntimeError("Resource not initialized yet")
        return self.__resource__

    @property
    def access(self) -> AccessType:
        """None: Access control rules for the entire resource."""
        return self.__access__

    @property
    def additional_metadata(self) -> List[Any]:
        """List[Any]: A flexible field for including any other relevant metadata."""
        return self.__additional__

    @property
    def annotations(self) -> List[EML.Annotation]:
        """Annotation: A list of precisely-defined semantic statements about this resource."""
        return self.__annotation__

    def initialize_resource(
            self,
            titles: str | I18nString | List[str | I18nString],
            creators: ResponsibleParty | List[ResponsibleParty],
            **kwargs
    ) -> None:
        """
        Initialize the resource instance of this EML.

        Parameters
        ----------
        titles : str | I18nString | List[str | I18nString]
            The titles to initialize the resource with.
        creators : ResponsibleParty | List[ResponsibleParty]+
            The creators to initialize the resource with.

        Other Parameters
        ----------------
        **kwargs : Parameters given in :class:`eml.resources.resource.Resource` or any subclass.
        """
        proper_titles = list()
        if isinstance(titles, list):
            for title in titles:
                proper_titles.append(I18nString(title))
        else:
            proper_titles.append(I18nString(titles))
        self.__resource__ = self.__resource_class__(
            titles=proper_titles,
            creators=creators if isinstance(creators, list) else [creators],
            **kwargs
        )
        return

    def add_title(self, title: str | I18nString, language: Language = Language.ENG) -> None:
        """
        Add a new title to the EML.

        Parameters
        ----------
        title : str | I18nString
            The title as string or with language already
        language : Language, optional
            The language of the title. If the title given is a I18nString, this argument is discarded.
        """
        self.resource.__titles__.append(I18nString(title, language))
        return

    def add_creator(self, creator: ResponsibleParty) -> None:
        """
        Add a new creator to the EML.

        Parameters
        ----------
        creator : ResponsibleParty
            The creator to be added to the resource.
        """
        self.resource.__creators__.append(creator)
        return

    def add_alternative_identifier(self, alternative_identifier: str | ExtensionString) -> None:
        """
        Add an alternative identifier to the EML.

        Parameters
        ----------
        alternative_identifier : str | ExtensionString
            An alternative identifier.
        """
        self.resource.__alternative_identifier__.append(ExtensionString(alternative_identifier))
        return

    def set_short_name(self, short_name: str) -> None:
        """
        Set short name of Resource in the EML.

        Parameters
        ----------
        short_name : str
            A short name for the resource.
        """
        self.resource.__short_name__ = short_name
        return

    def add_metadata_provider(self, metadata_provider: ResponsibleParty) -> None:
        """
        Add metadata provider to the EML.

        Parameters
        ----------
        metadata_provider : ResponsibleParty
            A Metadata Provider of the resource in the EML.
        """
        self.resource.__metadata_provider__.append(metadata_provider)
        return

    def add_associated_party(self, associated_party: ResponsibleParty, role: Role) -> None:
        """
        Add metadata provider to the EML.

        Parameters
        ----------
        associated_party : ResponsibleParty
            An Associated Party of the resource in the EML.
        role : Role
            The role of the Associated Party.
        """
        self.resource.__associated__.append((associated_party, role))
        return

    def set_pub_date(self, date: dt.date) -> None:
        """
        Set publication date of the resource in the EML.

        Parameters
        ----------
        date : datetime.date
            Publication date.
        """
        self.resource.__pub_date__ = date
        return

    def set_language(self, language: Language) -> None:
        """
        Set language of the resource in the EML.

        Parameters
        ----------
        language : Language
            A Language enum instance.
        """
        self.resource.__lang__ = language
        return

    def set_series(self, series: str) -> None:
        """
        Set series of the resource in the EML.

        Parameters
        ----------
        series : str
            The series from which the resource came.
        """
        self.resource.__series__ = series
        return

    def set_abstract(self, abstract: EMLTextType) -> None:
        """
        Set abstract of the resource in the EML.

        Parameters
        ----------
        abstract : EMLTextType
            An abstract in the required :class:`eml.types.text_type.EMLTextType` instance format.
        """
        self.resource.__abstract__ = abstract
        return

    def add_keyword_set(self, keyword_set: EMLKeywordSet) -> None:
        """
        Add keyword set to the resource in the EML.

        Parameters
        ----------
        keyword_set : EMLKeywordSet
            An instance of :class:`eml.resources.keyword_set.EMLKeywordSet`.
        """
        self.resource.__keywords__.append(keyword_set)
        return

    def add_additional_info(self, additional_info: EMLTextType) -> None:
        """
        Add additional info to the resource in the EML.

        Parameters
        ----------
        additional_info : EMLTextType
            Additional info in the required :class:`eml.types.text_type.EMLTextType` instance format.
        """
        self.resource.__add_info__.append(additional_info)
        return

    def set_intellectual_rights(self, intellectual_rights: EMLTextType) -> None:
        """
        Set intellectual rights of the resource in the EML.

        Parameters
        ----------
        intellectual_rights : EMLTextType
            The intellectual rights in the required :class:`eml.types.text_type.EMLTextType` instance format.
        """
        self.resource.__int_rights__ = intellectual_rights
        return

    def add_licensed(self, licensed: EMLLicense) -> None:
        """
        Add licensed to the resource in the EML.

        Parameters
        ----------
        licensed : EMLLicense
            An license instance.
        """
        self.resource.__licensed__.append(licensed)
        return

    def add_distribution(self, distribution: EMLDistribution) -> None:
        """
        Add distribution to the resource in the EML.

        Parameters
        ----------
        distribution : EMLDistribution
            An instance of :class:`eml.resources.distribution.distribution.EMLDistribution`.
        """
        self.resource.__dist__.append(distribution)
        return

    def set_coverage(self, coverage: EMLCoverage) -> None:
        """
        Set coverage of the resource in the EML.

        Parameters
        ----------
        coverage : EMLCoverage
            An instance of :class:`eml.resources.coverage.coverage.EMLCoverage`.
        """
        self.resource.__cover__ = coverage
        return

    def add_annotation(self, annotation: SemanticAnnotation) -> None:
        """
        Add annotation to the resource in the EML.

        Parameters
        ----------
        annotation : SemanticAnnotation
            An annotation in the :class:`eml.types.semantic_annotation.SemanticAnnotation` instance format.
        """
        self.resource.annotate(annotation)
        return

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLObject:
        """
        EML cannot have references.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance.
        nmap : Dict
            Namespace.

        Raises
        ------
        ValueError
            EML cannot reference another EML.
        """
        raise ValueError("EML cannot reference another EML")

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EML:
        """
        Generate an EML instance object using an XML element.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        EML
            EML instance.
        """
        assert element.get("packageId", None) is not None, "`packageId` attribute is not present in document"
        access = None
        access_elem = element.find("access", nmap)
        if access_elem is not None:
            access = AccessType.parse(access_elem, nmap)
        additional_metadata = list()
        for add_metadata in element.findall("additionalMetadata", nmap):
            additional_metadata.append(EMLMetadata.parse(add_metadata, nmap))
        annotation = list()
        annotations_elem = element.find("annotations", nmap)
        if annotations_elem is not None:
            for annotation_elem in annotations_elem.findall("annotation", nmap):
                annotation.append((
                    SemanticAnnotation.parse(annotation_elem, nmap),
                    annotation_elem.get("references", None)
                ))
        eml = EML(
            package_id=element.get("packageId"),
            system=element.get("system"),
            resource_type=EMLResource.get_resource_type(element),
            version=EMLVersion.get_version(element.get(f"{{{nmap['xsi']}}}schemaLocation", None)),
            language=Language.get_language(element.get(f"{{{nmap['xml']}}}lang", "eng")),
            access=access,
            additional_metadata=additional_metadata,
            annotation=annotation
        )
        eml.__resource__ = eml.__resource_class__.parse(
            element.find(eml.__resource_type__.name.lower()), nmap
        )
        eml.__namespace__ = nmap
        return eml

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance using the EML information.

        Returns
        -------
        :class:`lxml.etree.Element`
            XML element instance.
        """
        try:
            root = super().to_element()
        except KeyError:
            for key, value in self.__version__.get_namespace().items():
                self.__namespace__[key] = value
            root = super().to_element()
        root.set("packageId", self.package_id)
        root.set(f"{{{self.__namespace__['xsi']}}}schemaLocation", " ".join(self.__version__.schema_location()))
        root.set("system", self.system)
        root.set("scope", self.scope.name.lower())
        root.set(f"{{{self.__namespace__['xml']}}}lang", self.language.name.lower())
        root.append(self.resource.to_element())
        if self.access is not None:
            root.append(self.access.to_element())
        for add_meta in self.additional_metadata:
            root.append(add_meta.to_element())
        if len(self.annotations) > 0:
            annotations_elem = self.object_to_element("annotations")
            for annotation in self.annotations:
                annotation.annotation.set_tag("annotation")
                annotation_elem = annotation.annotation.to_element()
                annotation_elem.set("references", annotation.references)
                annotations_elem.append(annotation_elem)
            root.append(annotations_elem)
        return root

    def __str__(self) -> str:
        return f"EML:\n\tResource Type: {self.resource_type.name}\n{self.resource}"

    # TODO: Method to make sure elements added references existing ones.



import unittest

from datetime_interval import Interval
from lxml import etree as et
import datetime as dt

from dwca.terms import EventDate
from test_xml.test_xml import TestXML


class TestUnionDateTimeField(TestXML):
    def test_parse(self):
        event_date_xml = """<field index="3" term="http://rs.tdwg.org/dwc/terms/eventDate"/>"""
        event_date = EventDate.from_string(event_date_xml)
        self.assertEqual(
            "http://rs.tdwg.org/dwc/terms/eventDate",
            event_date.uri,
            "Wrong URI"
        )
        self.assertEqual(3, event_date.index, "Wrong index")
        self.assertEqualTree(
            et.fromstring(event_date_xml),
            event_date.to_element(),
            "Wrong generated element."
        )
        pass

    def test_format(self):
        event_date = EventDate(3)
        value = "1963-03-08T14:07-0600"
        self.assertEqual(
            dt.datetime(year=1963, month=3, day=8, hour=14, minute=7, tzinfo=dt.timezone(dt.timedelta(hours=-6))),
            event_date.format(value),
            "Error formatting union (datetime)"
        )
        value = "2007-03-01T13:00:00Z/2008-05-11T15:30:00Z"
        actual = event_date.format(value)
        expected = Interval(
            start=dt.datetime(year=2007, month=3, day=1, hour=13, tzinfo=dt.timezone.utc),
            end=dt.datetime(year=2008, month=5, day=11, hour=15, minute=30, tzinfo=dt.timezone.utc)
        )
        # self.assertEqual(expected, actual, "Error formatting union (interval)")
        self.assertEqual(
            expected.start, actual.start,
            "Error formatting union (interval, start)"
        )
        self.assertEqual(
            expected.end, actual.end,
            "Error formatting union (interval, end)"
        )
        value = "Invalid value"
        self.assertRaisesRegex(TypeError, "does not match any", event_date.format, value)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import os
import shutil
import sys
import tempfile
import warnings
from abc import ABC
from copy import deepcopy
from enum import Enum
from typing import List, Dict, Type, Tuple, BinaryIO, Generator, Any
from warnings import warn

from lxml import etree as et

from dwca.terms import Field, DWCType, DWCModified, DWCLanguage, DWCLicense, DWCRightsHolder, DWCAccessRights, \
    DWCBibliographicCitation, DWCReferences, DWCInstitution, DWCCollection, DWCDataset, DWCInstitutionCode, \
    DWCCollectionCode, DWCDatasetName, DWCOwnerInstitutionCode, DWCBasisOfRecord, DWCInformationWithheld, \
    DWCDataGeneralizations, DWCDynamicProperties, OutsideTerm
from xml_common import XMLObject
from xml_common.utils import iterate_with_bar, type_to_pl, format_to_sql

try:
    import pandas as pd
except Exception:
    pd = None
try:
    import polars as pl
except Exception:
    pl = None


class DataFileType(Enum):
    """
    Type of data file in the Darwin Core Archive.
    """
    CORE = 0
    EXTENSION = 1


class DataFile(XMLObject, ABC):
    """
    Abstract class representing the class of data represented by each row in a data entity.

    Parameters
    ----------
    _id : int
        Unique identifier for the core entity.
    files : str
        File location, in the archive, this is inside the `zip` file.
    fields : List[Field]
        A list of the Field (columns) in the Core data entity.
    data_file_type: DataFileType
        The Data File Type in the Darwin Core Archive.
    encoding : str, optional
        Encoding of the file location (`files` parameter), default is "utf-8".
    lines_terminated_by : str, optional
        Delimiter of lines on the file, default `"\\\\n"`.
    fields_terminated_by : str, optional
        Delimiter of the file (cells) on the file, default `","`.
    fields_enclosed_by : str, optional
        Specifies the character used to enclose (mark the start and end of) each field, default empty `""`.
    ignore_header_lines : int, optional
        Number of lines to ignore at the start of document. Default 0 lines.
    """
    URI = "http://rs.tdwg.org/dwc/terms/"
    """str: Unified Resource Identifier (URI) for the term identifying the class of data."""
    __field_class__ = [
        DWCType, DWCModified, DWCLanguage, DWCLicense, DWCRightsHolder,
        DWCAccessRights, DWCBibliographicCitation, DWCReferences,
        DWCInstitution, DWCCollection, DWCDataset, DWCInstitutionCode,
        DWCCollectionCode, DWCDatasetName, DWCOwnerInstitutionCode,
        DWCBasisOfRecord, DWCInformationWithheld, DWCDataGeneralizations,
        DWCDynamicProperties,
    ]

    class Entry:
        def __init__(self, **kwargs) -> None:
            for key, value in kwargs.items():
                setattr(self, key, value)
            return

        def to_dict(self) -> Dict:
            return self.__dict__

    def __init__(
            self, _id: int, files: str,
            fields: List[Field],
            data_file_type: DataFileType = DataFileType.CORE,
            encoding: str = "utf-8",
            lines_terminated_by: str = "\n",
            fields_terminated_by: str = ",",
            fields_enclosed_by: str = "",
            ignore_header_lines: int = 0,
    ) -> None:
        super().__init__()
        self.__id__ = _id
        self.__files__ = files
        self.__type__ = data_file_type
        self.__fields__ = fields
        self.__check_fields__()
        self.__encoding__ = encoding
        self.__lines_end__ = lines_terminated_by
        self.__fields_end__ = fields_terminated_by
        self.__fields_enclosed__ = fields_enclosed_by
        self.__ignore_header_lines__ = ignore_header_lines
        self.PRINCIPAL_TAG = self.__type__.name.lower()
        self.__entries__: List[DataFile.Entry] = list()
        self.__data__ = None
        self.__lazy__ = False
        self.__temp_file__ = ""
        self.__sql__ = ""
        self.__primary_key__ = None
        self.__core_field__ = self.__type__ == DataFileType.CORE
        self.__observers__: List[Tuple[int, DarwinCoreArchive]] = list()
        return

    @property
    def id(self) -> int:
        """int: Column to be identified as primary key"""
        return self.__id__

    @property
    def uri(self) -> str:
        """str: Unified Resource Identifier (URI) for the term."""
        return self.URI

    @property
    def filename(self) -> str:
        """str: Filename of the Data File entity."""
        return self.__files__

    def is_lazy(self) -> bool:
        """
        Check if data file load its data as a Lazy Frame.

        Returns
        -------
        bool
            True when file read as a Lazy Frame, False otherwise.
        """
        return self.__lazy__

    @property
    def fields(self) -> List[str]:
        """List[str]: List of terms of this data file."""
        return [field.uri for field in self.__fields__]

    @property
    def name(self) -> str:
        """str: The name of the field."""
        names = self.uri.split("/")
        if len(names) > 1:
            return names[-1]
        else:
            return names[0]

    def add_field(self, field: Field) -> None:
        """
        Add a field to this Data File.

        Parameters
        ----------
        field : Field
            A :class:`dwca.terms.field.Field` object.

        Returns
        -------
        None
        """
        if field.index != len(self.__fields__):
            warn(f"Field must be added at the end of current data, setting index to {len(self.__fields__)}")
            field.index = len(self.__fields__)
        field.__namespace__ = self.__namespace__
        self.__fields__.append(field)
        if self.is_lazy():
            self.__data__ = self.__data__.with_columns(
                pl.lit(
                    None if field.default is None else field.default
                ).alias(field.name)
            )
        elif len(self.__entries__) > 0:
            self.__data__ = None
            for entry in self.__entries__:
                setattr(entry, field.name, None if field.default is None else field.default)
        return

    @property
    def pandas(self) -> pd.DataFrame:
        """pandas.DataFrame: Data of this DataFile as pandas.DataFrame."""
        if self.__data__ is None:
            self.as_pandas()
        return self.__data__

    @pandas.setter
    def pandas(self, df: pd.DataFrame) -> None:
        assert len(self.__data__.columns) == len(df.columns)
        self.__data__ = df
        self.__entries__.clear()
        for _, row in df.iterrows():
            self.__entries__.append(DataFile.Entry(**row.to_dict()))
        for i, observer in self.__observers__:
            if self.__type__ == DataFileType.CORE:
                observer.core = self
            else:  # self.__type__ == DataFileType.EXTENSION
                observer.extensions[i] = self
        return

    @property
    def polars(self) -> pl.DataFrame:
        """polars.DataFrame: Data of this DataFile as polars.DataFrame."""
        if self.__data__ is None or self.is_lazy():
            self.as_polars()
        return self.__data__

    @property
    def sql_table(self) -> str:
        """str: Data file as CREATE TABLE sql statement."""
        if len(self.__sql__) == 0:
            self.generate_sql_table()
        return self.__sql__

    @property
    def insert_sql(self) -> Generator[str, Tuple[Any]]:
        """
        Generate the INSERT INTO sql statement and the values to be inserted.
        """
        if self.__type__ == DataFileType.EXTENSION and not self.__core_field__:
            raise RuntimeError("`set_core_field` must be called before `insert_sql` in Extension DataFile.")
        statement = f"INSERT INTO \"{self.name}\" (\n"
        columns = list()
        types = list()
        for i, field in enumerate(self.__fields__):
            columns.append(f"\"{field.name}\"")
            types.append("%s")
        statement += ",\n".join(columns)
        statement += "\n) VALUES ("
        statement += ", ".join(types)
        statement += ")\n"
        if self.is_lazy():
            for i in range(len(self)):
                tmp = self.__data__.slice(i, 1).collect()
                yield statement, tuple([
                    format_to_sql(tmp.select(field.name).item(), field.TYPE)
                    for field in self.__fields__
                ])
                del tmp
        for entry in self.__entries__:
            yield statement, tuple([
                format_to_sql(getattr(entry, field.name), field.TYPE)
                for field in self.__fields__
            ])

    def _register_darwin_core_(self, _on: int, dwca: DarwinCoreArchive) -> None:
        self.__observers__.append((_on, dwca))
        return

    def __len__(self) -> int:
        if self.__data__ is not None:
            try:
                return len(self.__data__)
            except TypeError:
                return self.__data__.select(pl.len()).collect().item() - self.__ignore_header_lines__
        else:
            return len(self.__entries__)

    @classmethod
    def get_term_class(cls, element: et.Element) -> Type[Field]:
        """
        Extract the Python ``class`` term from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element instance.

        Returns
        -------
        Type[Field]
            The Python ``class`` representing the term name.
        """
        for field_class in cls.__field_class__:
            if element.get("term") == field_class.URI:
                return field_class
        warn(f"{element.get('term')} not in expected namespace for "
             f"{cls.URI} class. "
             f"Some functionalities may not be available.")
        return OutsideTerm

    @classmethod
    def parse_kwargs(cls, element: et.Element, nmap: Dict) -> Dict:
        """
        Parse an `lxml.etree.Element` into the DataFile parameters.

        Parameters
        ----------
        element : `lxml.etree.Element`
            An XML `Element`.
        nmap : Dict
            Dictionary of prefix:uri.

        Returns
        -------
        Dict
            The Parameters of any DataFile.
        """
        fields = list()
        for field_tree in element.findall("field", namespaces=nmap):
            try:
                fields.append(cls.get_term_class(field_tree).parse(field_tree, nmap))
            except TypeError as e:
                warn(f"Error on field {et.tostring(field_tree)}: {e}", category=SyntaxWarning)
                pass
        assert len(fields) >= 1, "A Data File must contain at least one field"
        df_type = DataFileType[et.QName(element).localname.upper()]
        if df_type == DataFileType.CORE:
            element_id = element.find("id", nmap)
        else:  # df_type == DataFileType.EXTENSION:
            element_id = element.find("coreid", nmap)
        _id = int(element_id.get("index"))
        fields_enclosed_by = element.get("fieldsEnclosedBy", "")
        return {
            "_id": _id,
            "files": element.find("files", namespaces=nmap).find("location", namespaces=nmap).text,
            "fields": fields,
            "data_file_type": df_type,
            "encoding": element.get("encoding", "utf-8"),
            "lines_terminated_by": cls.__format_escape__(element.get("linesTerminatedBy", "\n")),
            "fields_terminated_by": cls.__format_escape__(element.get("fieldsTerminatedBy", ",")),
            "fields_enclosed_by": element.get("fieldsEnclosedBy", ""),
            "ignore_header_lines": int(element.get("ignoreHeaderLines", 0)),
        }

    @staticmethod
    def __format_escape__(characters: str) -> str:
        return characters.replace(
            "\\n", "\n"
        ).replace(
            "\\r", "\r"
        ).replace(
            "\\t", "\t"
        )

    @staticmethod
    def __unformat_escape__(characters: str) -> str:
        return characters.replace(
            "\n", "\\n"
        ).replace(
            "\r", "\\r"
        ).replace(
            "\t", "\\t"
        )

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> DataFile | None:
        """
        Parse an `lxml.etree.Element` into a concrete DataFile object.

        Parameters
        ----------
        element : `lxml.etree.Element`
            An XML `Element`.
        nmap : Dict
            Dictionary of prefix:uri.

        Returns
        -------
        DataFile
            An instance of a concrete DataFile class.
        """
        if element is None:
            return None
        kwargs = cls.parse_kwargs(element, nmap)
        data_file = cls(**kwargs)
        data_file.__namespace__ = nmap
        return data_file

    def to_element(self) -> et.Element:
        """
        Generate a XML `Element` instance

        Returns
        -------
        `lxml.etree.Element`
            XML `Element` instance
        """
        element = super().to_element()
        element.set("rowType", self.uri)
        element.set("encoding", self.__encoding__)
        element.set("linesTerminatedBy", self.__unformat_escape__(self.__lines_end__))
        element.set("fieldsTerminatedBy", self.__unformat_escape__(self.__fields_end__))
        element.set("fieldsEnclosedBy", self.__fields_enclosed__)
        element.set("ignoreHeaderLines", str(self.__ignore_header_lines__))
        if self.__type__ == DataFileType.CORE:
            element_id = self.object_to_element("id")
        else:  # self.__type__ == DataFileType.EXTENSION:
            element_id = self.object_to_element("coreid")
        element_id.set("index", str(self.id))
        element.append(element_id)
        files = self.object_to_element("files")
        location = self.object_to_element("location")
        location.text = self.filename
        files.append(location)
        element.append(files)
        for field in self.__fields__:
            if self.__type__ == DataFileType.EXTENSION and field.index == self.id:
                continue
            field_element = field.to_element()
            element.append(field_element)
        return element

    @classmethod
    def check_principal_tag(cls, tag: str, nmap: Dict) -> None:
        """
        Overwrite due to different possible tags.

        Parameters
        ----------
        tag : str
            Actual tag.
        nmap : Dict
            Namespace.
        """
        pass

    def __check_fields__(self) -> None:
        extension = self.__type__ == DataFileType.EXTENSION
        ordered_fields: List[Field] = [None] * (len(self.__fields__) + extension)
        for field in self.__fields__:
            try:
                if ordered_fields[field.index] is not None:
                    raise AssertionError(f'"index={field.index}" declared twice.')
                ordered_fields[field.index] = field
            except IndexError:
                pass
        if extension:
            if ordered_fields[self.id] is None:
                ordered_fields[self.id] = OutsideTerm(self.id, "")
            else:
                ordered_fields.pop()
        nones = [str(i) for i, item in enumerate(ordered_fields) if item is None]
        if len(nones) > 0:
            raise AssertionError(f'Index {"".join(nones)} not declared.')
        self.__fields__ = ordered_fields
        return

    def read_file(self, content: str, source_file: BinaryIO = None, lazy: bool = False) -> None:
        """
        Read the content of the file specified in `files` parameters (:meth:`filename`).

        Parameters
        ----------
        content : str
            Content of the file
        source_file : BinaryIO, optional
            File to read in case of laziness.
        lazy : bool, optional
            Read the file in lazy evaluation mode. Default `False`.
        """
        if lazy:
            try:
                import polars as pl
                with tempfile.NamedTemporaryFile(delete=False) as file:
                    shutil.copyfileobj(source_file, file)
                    self.__data__ = pl.scan_csv(
                        file.name,
                        has_header=False,
                        skip_rows=self.__ignore_header_lines__,
                        separator=self.__fields_end__,
                        quote_char=None,
                        schema={field.name: type_to_pl(field.TYPE, lazy=True) for field in self.__fields__},
                    )
                    self.__lazy__ = True
                    self.__temp_file__ = file.name
                    warn("Reading in lazy evaluation mode generates a temporal file, make sure to call close() to "
                         "delete it")
            except ImportError:
                raise ImportError("Cannot read lazy without polars installed.")
        else:
            lines = content.split(self.__lines_end__)
            lines = list(filter(lambda x: x != "", lines))
            for line in iterate_with_bar(
                    lines[self.__ignore_header_lines__:],
                    desc=f"Reading file {self.filename}", unit="entry"
            ):
                kwargs = dict()
                for field, value in zip(self.__fields__, line.split(self.__fields_end__)):
                    kwargs[field.name] = field.format(value)
                self.__entries__.append(DataFile.Entry(**kwargs))
        return


    def write_file(self) -> str:
        """
        Write the content as a text using format information on this object.

        Returns
        -------
        str
            Data File as plain text.
        """
        output_file = ""
        header = [field.name for field in self.__fields__]
        output_file += f"{self.__fields_end__}".join(header) + self.__lines_end__
        for entry in iterate_with_bar(self.__entries__, desc=f"Writing data {self.uri}", unit="line"):
            line = list()
            for field in self.__fields__:
                try:
                    line.append(field.unformat(getattr(entry, field.name)))
                except AssertionError:  # In case the type got lost in pandas
                    line.append(field.unformat(field.TYPE(getattr(entry, field.name))))
                except Exception as e:
                    print(f"Error on {field.name} with value {getattr(entry, field.name)}", file=sys.stderr)
                    raise e
            output_file += f"{self.__fields_end__}".join(line) + self.__lines_end__
        return output_file

    def as_pandas(self, _no_interaction: bool = False) -> pd.DataFrame:
        """
        Convert information in this DataFile in a pandas.DataFrame.

        Returns
        -------
        DataFrame
            Information as a pandas.DataFrame.
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("Install pandas to use this feature.")
        fields = list()
        for field in self.__fields__:
            fields.append(field.name)
        entries = list()
        if _no_interaction:
            for entry in self.__entries__:
                entries.append(entry.to_dict())
        else:
            for entry in iterate_with_bar(self.__entries__, desc="Converting to pandas", unit="entry"):
                entries.append(entry.to_dict())
        self.__data__ = pd.DataFrame(entries, columns=fields)
        return self.__data__

    def as_polars(self, _no_interaction: bool = False) -> pl.DataFrame:
        """
        Convert information in this DataFile in a polars DataFrame.

        Returns
        -------
        DataFrame:
            Information as a polars.DataFrame.
        """
        try:
            import polars as pl
        except ImportError:
            raise ImportError("Install polars to use this feature.")
        if self.is_lazy():
            self.__data__ = self.__data__.collect()
            self.__lazy__ = False
            self.close()
            return self.__data__
        else:
            fields = list()
            for field in self.__fields__:
                fields.append((field.name, type_to_pl(field.TYPE)))
            entries = list()
            if _no_interaction:
                for entry in self.__entries__:
                    entries.append(entry.to_dict())
            else:
                for entry in iterate_with_bar(self.__entries__, desc="Converting to polars", unit="entry"):
                    entries.append(entry.to_dict())
            self.__data__ = pl.DataFrame(entries, schema=fields)
            return self.__data__

    def set_core_field(self, field: Field) -> None:
        """
        Set the Core field in an Extension DataFile.

        Parameters
        ----------
        field : Field
            Primary Key from the Core DataFile.

        Returns
        -------
        None
        """
        if self.__type__ == DataFileType.CORE:
            raise AttributeError(f"Core DataField cannot change primary field.")
        self.__fields__[self.id] = field
        self.__fields__[self.id].__index__ = self.id
        self.__core_field__ = True
        return

    def set_primary_key(self, primary_key: str) -> None:
        """
        Set the primary key in an Extension DataFile to be referenced on the new SQL table.

        Parameters
        ----------
        primary_key : str
            Name of the Core Data File.

        Returns
        -------
        None
        """
        if self.__type__ == DataFileType.CORE:
            raise AttributeError(f"Core DataField cannot be set the primary key.")
        self.__primary_key__ = primary_key
        return

    def generate_sql_table(self) -> str:
        """
        Generate the CREATE TABLE statement for SQL database.

        Returns
        -------
        str
            CREATE TABLE statement.
        """
        if self.__type__ == DataFileType.EXTENSION and self.__primary_key__ is None:
            raise RuntimeError("`set_primary_key` must be called before `generate_sql_table` in Extension DataFile.")
        sql_columns = ""
        for field in self.__fields__:
            sql_columns += f"\"{field.name}\" {field.sql_type},\n"
        primary_key = self.__fields__[self.id].name
        if primary_key == "":
            raise RuntimeError("Primary key cannot be empty, in case of Extension, called `set_core_field` first.")
        foreign_key = ""
        if self.__type__ == DataFileType.EXTENSION:
            foreign_key = f""",
            FOREIGN KEY ("{primary_key}") REFERENCES \"{self.__primary_key__}\""""
        self.__sql__ = f"""CREATE TABLE "{self.name}" (
            {sql_columns}
            PRIMARY KEY ("{primary_key}"){foreign_key}
        );"""
        return self.__sql__

    def merge(self, data_file: DataFile) -> DataFile:
        assert self.uri == data_file.uri, "Cannot merge two different classes: `{}` and `{}`".format(
            self.uri, data_file.uri
        )
        merged = deepcopy(self)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            for field in data_file.__fields__:
                if field.uri not in merged.fields:
                    merged.add_field(field)
        try:
            data = merged.pandas
            merged.pandas = pd.concat([data, data_file.pandas], axis=0).reset_index(drop=True)
        except ImportError:
            for entry in data_file.__entries__:
                merged.__entries__.append(entry)
            for i, entry in enumerate(merged.__entries__):
                for field in merged.__fields__:
                    try:
                        getattr(entry, field.name)
                    except AttributeError:
                        setattr(merged.__entries__[i], field.name, field.default)
        return merged

    def close(self) -> None:
        if self.is_lazy():
            os.remove(self.__temp_file__)
            self.__temp_file__ = ""
            self.__lazy__ = False
        return

    def __str__(self) -> str:
        role = self.__type__.name.lower().capitalize()
        return (f"{role}:"
                f"\n\tclass: {self.uri}"
                f"\n\tfilename: {self.filename}"
                f"\n\tcontent: {self.__len__()} entries")



import unittest

from eml import EML
from eml.types import SemanticAnnotation
from test_xml.test_xml import TestXML


class TestAnnotation(TestXML):
    def setUp(self) -> None:
        text_xml = """
<annotation id="Annotation">
    <propertyURI 
        label="has characteristic"
    >http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#ofCharacteristic</propertyURI>
    <valueURI
        label="Mass"
    >http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#Mass</valueURI>
</annotation>
        """
        self.semantic_annotation = SemanticAnnotation.from_string(text_xml)
        return

    def test_equal(self):
        first_annotation = EML.Annotation(self.semantic_annotation, "example")
        self.assertIn(
            self.semantic_annotation,
            (first_annotation, ),
            "Semantic annotation not the same"
        )
        self.assertEqual(
            first_annotation,
            first_annotation,
            "Same annotation not equal"
        )
        second_annotation = EML.Annotation(self.semantic_annotation, "example")
        self.assertEqual(
            first_annotation,
            second_annotation,
            "Equal annotation not equal"
        )
        self.assertNotEqual(1, first_annotation, "Annotation equal to an int")

    def test_annotation(self):
        annotation = EML.Annotation(self.semantic_annotation, "example")
        self.assertEqual(
            self.semantic_annotation,
            annotation.annotation,
            "Annotation incorrectly got"
        )

    def test_reference(self):
        annotation = EML.Annotation(self.semantic_annotation, "example")
        self.assertEqual(
            "example",
            annotation.references,
            "References incorrectly got"
        )


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from lxml import etree as et

from dwca.base import DarwinCoreArchive
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestMeta(TestXML):
    def setUp(self) -> None:
        with open(os.path.join(PATH, os.pardir, "example_data", "meta.xml"), "r", encoding="utf-8") as file:
            content = file.read()
        self.metadata = et.fromstring(content)
        self.text = et.tostring(self.metadata, pretty_print=True).decode("utf-8")
        return

    def test_parse(self):
        meta = DarwinCoreArchive.Metadata.from_string(self.text)
        self.assertEqual("eml.xml", meta.__metadata__,"Metadata not found in file")
        self.assertEqual(0, meta.__core__.id, "Core not found")
        self.assertEqual(3, len(meta.__extensions__), "Extensions not found")
        self.assertEqual(0, meta.__extensions__[0].id, "First extension not found")
        self.assertEqual(0, meta.__extensions__[1].id, "Second extension not found")
        self.assertEqual(0, meta.__extensions__[2].id, "Third extension not found")
        self.assertEqualTree(self.metadata, meta.to_element(), "Not same tree generated")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from enum import Enum


class CamelCaseEnum(Enum):
    """
    Enum class that allow conversion and retrieval using camel case name.
    """
    def to_camel_case(self):
        """
        Converts name to camel case (eg: camelCase).

        Returns
        -------
        str
            Name of role in camel case.
        """
        words = self.name.split("_")
        return "".join([
            word.lower() if i == 0 else word.lower().capitalize()
            for i, word in enumerate(words)
        ])

    def __str__(self) -> str:
        words = self.name.split("_")
        return " ".join([
            word.capitalize() for word in words
        ])

    @classmethod
    def get_enum(cls, name: str) -> CamelCaseEnum:
        """
        Obtain the Enum from the string with the name in camel case.

        Parameters
        ----------
        name : str
            Camel case name.

        Returns
        -------
        CamelCaseEnum
            Instance of CamelCaseEnum.

        Raises
        ------
        ValueError
            Not a valid name.
        """
        for enum in cls:
            if enum.to_camel_case() == name:
                return enum
        raise ValueError(f"{name} not a valid name for {cls.__name__}.")



from eml.resources.coverage.geo_coverage import GeographicCoverage
from eml.resources.coverage.time_coverage import TemporalCoverage
from eml.resources.coverage.taxa_coverage import TaxonomicCoverage
from eml.resources.coverage.coverage import EMLCoverage



import unittest

from lxml import etree as et

from eml.types import AccessType, Scope, AccessPermission, AccessRole
from test_xml.test_xml import TestXML


class TestAccessType(TestXML):
    DEFAULT_TAGS = {
        "order": "allowFirst",
        "scope": "document",
    }

    def test_parse(self):
        text_xml = """
<access authSystem="ldap://ldap.ecoinformatics.org:389/dc=ecoinformatics,dc=org">
    <allow>
        <principal>uid=alice,o=LTER,dc=ecoinformatics,dc=org</principal>
        <permission>read</permission>
        <permission>write</permission>
    </allow>
    <deny>
        <principal>public</principal>
        <permission>write</permission>
    </deny>
</access>
        """
        access = AccessType.from_string(text_xml)
        self.assertEqual(
            "ldap://ldap.ecoinformatics.org:389/dc=ecoinformatics,dc=org",
            access.auth_system,
            "Error on parsing auth system"
        )
        self.assertFalse(access.referencing, "Error on parsing referencing")
        self.assertEqual(1, len(access.allow.principal), "Error on parsing allow (principal)")
        self.assertEqual(
            "uid=alice,o=LTER,dc=ecoinformatics,dc=org",
            access.allow.principal[0],
            "Error on parsing allow (principal)"
        )
        self.assertEqual(2, len(access.allow.permission), "Error on parsing allow (permission)")
        self.assertEqual(
            AccessPermission.READ,
            access.allow.permission[0],
            "Error on parsing allow (permission)"
        )
        self.assertEqual(
            AccessPermission.WRITE,
            access.allow.permission[1],
            "Error on parsing allow (permission)"
        )
        self.assertEqual(1, len(access.deny.principal), "Error on parsing deny (principal)")
        self.assertEqual(
            "public",
            access.deny.principal[0],
            "Error on parsing deny (principal)"
        )
        self.assertEqual(1, len(access.deny.permission), "Error on parsing deny (permission)")
        self.assertEqual(
            AccessPermission.WRITE,
            access.deny.permission[0],
            "Error on parsing deny (permission)"
        )
        self.assertEqualTree(et.fromstring(text_xml), access.to_element(), "Error on to element")

    def test_parse_none(self):
        self.assertIsNone(AccessType.parse(None, {}), "Element from nowhere")
        self.assertIsNone(AccessRole.parse(None, {}), "Element from nowhere")

    def test_referencing(self):
        text_xml = """
<access authSystem="ldap://ldap.ecoinformatics.org:389/dc=ecoinformatics,dc=org">
    <references system="http://gbif.org">Access Id</references>
</access>
        """
        access = AccessType.from_string(text_xml)
        self.assertEqual(
            "ldap://ldap.ecoinformatics.org:389/dc=ecoinformatics,dc=org",
            access.auth_system,
            "Error on parsing auth system"
        )
        self.assertEqual("Access Id", access.id, "Error on parsing id")
        self.assertEqual(Scope.DOCUMENT, access.scope, "Error on default scope")
        self.assertIsNone(access.system, "System from nowhere")
        self.assertTrue(access.referencing, "Error on parsing referencing")
        self.assertEqual("http://gbif.org", access.references.system, "Error on parsing references system")
        self.assertEqualTree(et.fromstring(text_xml), access.to_element(), "Error on to element")

    def test_error_initialization(self):
        self.assertRaises(TypeError, AccessType, "A system", order=45)

    def test_error_role_initialization(self):
        self.assertRaises(ValueError, AccessRole, [], [])
        self.assertRaises(ValueError, AccessRole, ["Control"], [])


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Dict, Union, List

from lxml import etree as et

from xml_common.utils import CamelCaseEnum
from eml.types import EMLObject, Scope, _NoTagObject


class AccessOrder(CamelCaseEnum):
    ALLOW_FIRST = 0
    DENY_FIRST = 1


class AccessPermission(CamelCaseEnum):
    READ = 0
    WRITE = 1
    CHANGE_PERMISSION = 2
    ALL = 3


class AccessRole(_NoTagObject):
    """
    Access Rules define a user's access to a resource.

    Parameters
    ----------
    principal : List[str]
        The user or group (principal) for which the access control applies.
    permission: List[AccessPermission]
         The type of permission being granted or denied.
    """
    def __init__(self, principal: List[str], permission: List[AccessPermission]) -> None:
        super().__init__()
        if len(principal) == 0:
            raise ValueError("At least one principal must be given")
        self.__principal__ = principal
        if len(permission) == 0:
            raise ValueError("At least one permission must be given")
        self.__permission__ = permission
        return

    @property
    def principal(self) -> List[str]:
        """
        List[str]: The user or group (principal) for which the access control applies.
        """
        return self.__principal__

    @property
    def permission(self) -> List[AccessPermission]:
        """List[AccessPermission]: The type of permission being granted or denied."""
        return self.__permission__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> AccessRole | None:
        """
        Generate a AccessRole instance from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance.
        nmap : Dict
            Namespace.

        Returns
        -------
        AccessRole
            An instance of AccessRole.
        """
        if element is None:
            return None
        principal = list()
        for prin_elem in element.findall("principal"):
            principal.append(prin_elem.text)
        permission = list()
        for perm_elem in element.findall("permission"):
            permission.append(AccessPermission.get_enum(perm_elem.text))
        access_role = AccessRole(principal, permission)
        access_role.__namespace__ = nmap
        return access_role

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance using this Access Role object.

        Returns
        -------
        lxml.etree.Element
            An XML element instance.
        """
        element = super().to_element()
        for principal in self.principal:
            prin_elem = self.object_to_element("principal")
            prin_elem.text = principal
            element.append(prin_elem)
        for permission in self.permission:
            perm_elem = self.object_to_element("permission")
            perm_elem.text = permission.to_camel_case()
            element.append(perm_elem)
        return element


class AccessType(EMLObject):
    """
    Access control rules for the entire resource, which can be overridden by access rules in distribution trees.

    Parameters
    ----------
    auth_system : str
        The authentication system is used to verify the user or group to whom access is allowed or denied.
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    order: Order, optional
        The order in which the allow and deny rules should be applied.
    allow: AccessRole
        A rule that grants a permission type. Mandatory if AccessType is not a reference.
    deny: AccessRole
        A rule that revokes a permission type. Mandatory if AccessType is not a reference.
    """
    PRINCIPAL_TAG = "access"

    def __init__(
            self, auth_system: str,
            _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
            referencing: bool = False,
            references_system: str = None,
            order: Union[AccessOrder, str] = AccessOrder.ALLOW_FIRST,
            allow: AccessRole = None,
            deny: AccessRole = None,
    ) -> None:
        super().__init__(_id, scope, system, referencing, references_system)
        self.__auth__ = auth_system
        if isinstance(order, AccessOrder):
            self.__order__ = order
        elif isinstance(order, str):
            self.__order__ = AccessOrder.get_enum(order)
        else:
            raise TypeError(f"'order' parameter must be a AccessOrder or sting, not a {type(order)}")
        self.__allow__ = allow
        self.__deny__ = deny
        return

    @property
    def auth_system(self) -> str:
        """str: The authentication system is used to verify the user or group to whom access is allowed or denied."""
        return self.__auth__

    @property
    def order(self) -> AccessOrder:
        """AccessOrder: The order in which the allow and deny rules should be applied."""
        return self.__order__

    @property
    def allow(self) -> AccessRole:
        """AccessRole: A rule that grants a permission type."""
        return self.__allow__

    @property
    def deny(self) -> AccessRole:
        """AccessRole: A rule that revokes a permission type."""
        return self.__deny__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> AccessType:
        """
        Generate an AccessType object that references another AccessType.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        AccessType
            Instance of AccessType
        """
        references = element.find("references", nmap)
        return AccessType(
            auth_system=element.get("authSystem"),
            _id=references.text,
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None)
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLObject:
        """
        Generate an AccessType object that does not reference another AccessType.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        AccessType
            Instance of AccessType
        """
        return AccessType(
            auth_system=element.get("authSystem"),
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=False,
            allow=AccessRole.parse(element.find("allow", nmap), nmap),
            deny=AccessRole.parse(element.find("deny", nmap), nmap),
        )

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance using this object.

        Returns
        -------
        lxml.etree.Element
            An XML element instance.
        """
        element = super().to_element()
        element = self._to_element_(element)
        element.set("authSystem", self.auth_system)
        element.set("order", self.order.to_camel_case())
        if self.referencing:
            return element
        self.allow.set_tag("allow")
        element.append(self.allow.to_element())
        self.deny.set_tag("deny")
        element.append(self.deny.to_element())
        return element



from __future__ import annotations

from typing import Dict, Any
from urllib.parse import urlparse

from lxml import etree as et

from eml.types import _NoTagObject


class ExtensionURI(_NoTagObject):
    """
    Extension URI representing an uri value with extra attributes.

    Parameters
    ----------
    value : str
        Value of the URI.
    label : str
        A human-readable representation of the URI.
    """
    def __init__(self, value: str, label: str) -> None:
        super().__init__()
        validate_url = urlparse(value)
        if validate_url.scheme == '' or validate_url.netloc == '':
            raise ValueError(f"{value} is not a valid URI")
        self.__value__ = value
        self.__label__ = label
        return

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ExtensionURI):
            return str(self) == str(other) and self.label == other.label
        elif isinstance(other, str):
            return str(self) == other
        else:
            return False

    def __str__(self) -> str:
        return self.__value__

    def __repr__(self) -> str:
        return f"<ExtensionURI ({str(self)} [label={self.label}])>"

    @property
    def label(self) -> str:
        """str: A human-readable representation of the URI."""
        return self.__label__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> ExtensionURI | None:
        """
        Generate an ExtensionURI object from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance.
        nmap : Dict
            Namespace.

        Returns
        -------
        ExtensionURI
            An URI with a human-readable label.
        """
        if element is None:
            return None
        return ExtensionURI(element.text, element.get("label"))

    def to_element(self) -> et.Element:
        """
        Generate an XML element with the information of the extension URI.

        Returns
        -------
        lxml.etree.Element
            An XML element instance.
        """
        element = super().to_element()
        element.set("label", self.label)
        element.text = str(self)
        return element
